/***************************************************************************
 *            burn.c
 *
 *  ven mar  3 18:50:18 2006
 *  Copyright  2006  Rouquier Philippe
 *  bonfire-app@wanadoo.fr
 ***************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include <glib-object.h>
#include <glib/gi18n-lib.h>

#include <libgnomevfs/gnome-vfs.h>

#include <nautilus-burn-drive.h>

#include "bonfire-marshal.h"
#include "burn-basics.h"
#include "burn.h"
#include "burn-job.h"
#include "burn-imager.h"
#include "burn-recorder.h"
#include "burn-readcd.h"
#include "burn-transcode.h"
#include "burn-mkisofs-base.h"
#include "burn-mkisofs.h"
#include "burn-cdrdao.h"
#include "burn-cdrecord.h"
#include "burn-growisofs.h"
#include "burn-dvd-rw-format.h"
#include "progress.h"
#include "burn-common.h"

static void bonfire_burn_class_init (BonfireBurnClass *klass);
static void bonfire_burn_init (BonfireBurn *sp);
static void bonfire_burn_finalize (GObject *object);

struct BonfireBurnPrivate {
	NautilusBurnDrive *drive;
	BonfireRecorder *recorder;
	BonfireImager *imager;

	gint64 image_size;

	BonfireBurnProgress *progress;
};

#define IS_LOCKED	"LOCKED"

typedef enum {
	ASK_DISABLE_JOLIET_SIGNAL,
	WARN_DATA_LOSS_SIGNAL,
	WARN_REWRITABLE_SIGNAL,
	INSERT_MEDIA_REQUEST_SIGNAL,
	PROGRESS_CHANGED_SIGNAL,
	ACTION_CHANGED_SIGNAL,
	LAST_SIGNAL
} BonfireBurnSignalType;

static guint bonfire_burn_signals [LAST_SIGNAL] = { 0 };
static GObjectClass *parent_class = NULL;

GType
bonfire_burn_get_type ()
{
	static GType type = 0;

	if(type == 0) {
		static const GTypeInfo our_info = {
			sizeof (BonfireBurnClass),
			NULL,
			NULL,
			(GClassInitFunc)bonfire_burn_class_init,
			NULL,
			NULL,
			sizeof (BonfireBurn),
			0,
			(GInstanceInitFunc)bonfire_burn_init,
		};

		type = g_type_register_static (G_TYPE_OBJECT, 
					       "BonfireBurn", 
					       &our_info,
					       0);
	}

	return type;
}

static void
bonfire_burn_class_init (BonfireBurnClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);

	parent_class = g_type_class_peek_parent(klass);
	object_class->finalize = bonfire_burn_finalize;
	
	bonfire_burn_signals [ASK_DISABLE_JOLIET_SIGNAL] =
		g_signal_new ("disable_joliet",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (BonfireBurnClass,
					       ask_disable_joliet),
			      NULL, NULL,
			      bonfire_marshal_INT__VOID,
			      G_TYPE_INT, 0);
        bonfire_burn_signals [WARN_DATA_LOSS_SIGNAL] =
		g_signal_new ("warn_data_loss",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (BonfireBurnClass,
					       warn_data_loss),
			      NULL, NULL,
			      bonfire_marshal_INT__VOID,
			      G_TYPE_INT, 0);
        bonfire_burn_signals [WARN_REWRITABLE_SIGNAL] =
		g_signal_new ("warn_rewritable",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (BonfireBurnClass,
					       warn_rewritable),
			      NULL, NULL,
			      bonfire_marshal_INT__VOID,
			      G_TYPE_INT, 0);
        bonfire_burn_signals [INSERT_MEDIA_REQUEST_SIGNAL] =
		g_signal_new ("insert_media",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (BonfireBurnClass,
					       insert_media_request),
			      NULL, NULL,
			      bonfire_marshal_INT__INT_INT,
			      G_TYPE_INT, 
			      2,
			      G_TYPE_INT,
			      G_TYPE_INT);
        bonfire_burn_signals [PROGRESS_CHANGED_SIGNAL] =
		g_signal_new ("progress_changed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (BonfireBurnClass,
					       progress_changed),
			      NULL, NULL,
			      bonfire_marshal_VOID__DOUBLE_LONG,
			      G_TYPE_NONE, 
			      2,
			      G_TYPE_DOUBLE,
			      G_TYPE_LONG);
        bonfire_burn_signals [ACTION_CHANGED_SIGNAL] =
		g_signal_new ("action_changed",
			      G_TYPE_FROM_CLASS (klass),
			      G_SIGNAL_RUN_LAST,
			      G_STRUCT_OFFSET (BonfireBurnClass,
					       action_changed),
			      NULL, NULL,
			      g_cclosure_marshal_VOID__INT,
			      G_TYPE_NONE, 
			      1,
			      G_TYPE_INT);
}

static void
bonfire_burn_init (BonfireBurn *obj)
{
	obj->priv = g_new0(BonfireBurnPrivate, 1);
}

static void
bonfire_burn_finalize (GObject *object)
{
	BonfireBurn *cobj;
	cobj = BONFIRE_BURN(object);

	if (cobj->priv->recorder)
		g_object_unref (cobj->priv->recorder);
	if (cobj->priv->imager)
		g_object_unref (cobj->priv->imager);
	if (cobj->priv->progress)
		g_object_unref (cobj->priv->progress);

	g_free (cobj->priv);
	G_OBJECT_CLASS(parent_class)->finalize(object);
}

BonfireBurn *
bonfire_burn_new ()
{
	BonfireBurn *obj;
	
	obj = BONFIRE_BURN (g_object_new (BONFIRE_TYPE_BURN, NULL));
	
	return obj;
}

static void
bonfire_burn_progress_changed (BonfireJob *job,
			       double fraction,
			       long time_remaining,
			       BonfireBurn *burn)
{
	double progress;

	if (fraction == -1.0) {
		gint64 isosize = -1;
		gint64 bytes_written = -1;

		/* sometimes when burning on the fly the recorder doesn't know
		 * in advance the size of the image it's burning */
		if (burn->priv->image_size)
			isosize = burn->priv->image_size;
		else if (burn->priv->imager)
			bonfire_imager_get_size (burn->priv->imager,
						 &isosize,
						 FALSE,
						 NULL);

		if (burn->priv->recorder
		&&  bonfire_job_is_running (BONFIRE_JOB (burn->priv->recorder)))
			bonfire_recorder_get_status (burn->priv->recorder,
						     NULL,
						     NULL,
						     &bytes_written);

		if (isosize > 0 && bytes_written > 0)
			fraction = (double) bytes_written / isosize;
		else
			fraction = 0.0;
	}

	if (burn->priv->recorder) {
		if (burn->priv->imager
		&& !bonfire_job_is_running (BONFIRE_JOB (burn->priv->imager))
		&&  BONFIRE_JOB (burn->priv->recorder) != BONFIRE_JOB (burn->priv->imager)
		&& (BONFIRE_JOB (burn->priv->recorder) == job))
			progress = (fraction + 1.0) / 2.0;
		else
			progress = fraction;
	}
	/* this is when an imager has a slave running (for example downloading) */
	else if (burn->priv->imager
	      && !bonfire_job_is_running (BONFIRE_JOB (burn->priv->imager)))
		progress = fraction;
	else if (burn->priv->drive->type != NAUTILUS_BURN_DRIVE_TYPE_FILE)
		progress = fraction / 2.0;
	else
		progress = fraction;

	if (burn->priv->progress) {
		gint64 b_written = -1;
		gint64 isosize = -1;
		int mb_isosize = -1;
		int mb_written = -1;
		int speed = -1;
		int fifo = -1;

		if (burn->priv->image_size) {
			isosize = burn->priv->image_size;
		}
		else if (burn->priv->imager) {
			bonfire_imager_get_size (burn->priv->imager,
						 &isosize,
						 FALSE,
						 NULL);
		}

		mb_isosize = (int) (isosize / 1048576);
		mb_written = (int) mb_isosize * fraction;

		if (burn->priv->recorder
		&&  bonfire_job_is_running (BONFIRE_JOB (burn->priv->recorder))) {
			bonfire_recorder_get_status (burn->priv->recorder,
						     &speed,
						     &fifo,
						     &b_written);
			if (!mb_written)
				mb_written = (int) b_written / 1048576;
		}
		else if (burn->priv->imager
		      &&  bonfire_job_is_running (BONFIRE_JOB (burn->priv->imager)))
			bonfire_imager_get_speed (burn->priv->imager,
						  &speed);

		/* FIXME: with DVDs we'd better write speed in DVD speed not CD */
		bonfire_burn_progress_set_status (burn->priv->progress,
						  progress,
						  time_remaining,
						  mb_isosize,
						  mb_written,
						  speed,
						  fifo);
	}
						  
	g_signal_emit (burn,
		       bonfire_burn_signals [PROGRESS_CHANGED_SIGNAL],
		       0,
		       progress,
		       time_remaining);
}

static void
bonfire_burn_action_changed_real (BonfireBurn *burn,
				  BonfireBurnAction action,
				  BonfireJob *job)
{
	if (burn->priv->progress) {
		char *string = NULL;

		if (action == BONFIRE_BURN_ACTION_FINISHED) {
			string = g_strdup (bonfire_burn_action_to_string (action));
		}
		else if (BONFIRE_JOB (burn->priv->imager) == job) {
			bonfire_job_get_action_string (BONFIRE_JOB (burn->priv->imager),
						       action,
						       &string);
		}
		else if (BONFIRE_JOB (burn->priv->recorder) == job) {
			bonfire_job_get_action_string (BONFIRE_JOB (burn->priv->recorder),
						       action,
						       &string);
		}

		bonfire_burn_progress_set_action (burn->priv->progress,
						  action,
						  string);

		if (string)
			g_free (string);
	}

	g_signal_emit (burn,
		       bonfire_burn_signals [ACTION_CHANGED_SIGNAL],
		       0,
		       action);
}

static void
bonfire_burn_action_changed (BonfireJob *job,
			     BonfireBurnAction action,
			     BonfireBurn *burn)
{
	bonfire_burn_action_changed_real (burn, action, job);
}

BonfireBurnResult
bonfire_burn_status (BonfireBurn *burn,
		     gint64 *isosize,
		     gint64 *written,
		     gint *speed,
		     gint *fifo)
{
	if (burn->priv->recorder
	&&  bonfire_job_is_running (BONFIRE_JOB (burn->priv->recorder))) {
		bonfire_recorder_get_status (burn->priv->recorder,
					     speed,
					     fifo,
					     written);
		if (burn->priv->imager)
			bonfire_imager_get_size (burn->priv->imager,
						 isosize,
						 FALSE,
						 NULL);
	}
	else if (burn->priv->imager) {
		if (burn->priv->image_size)
			*isosize = burn->priv->image_size;
		else
			bonfire_imager_get_size (burn->priv->imager,
						 isosize,
						 FALSE,
						 NULL);

		if (bonfire_job_is_running (BONFIRE_JOB (burn->priv->imager)))
			bonfire_imager_get_speed (burn->priv->imager, speed);
	}
	else
		return BONFIRE_BURN_NOT_READY;

	return BONFIRE_BURN_OK;
}

static void
bonfire_burn_check_media (BonfireBurn *burn,
			  NautilusBurnDrive *drive,
			  NautilusBurnMediaType *type,
			  gboolean *is_rewritable,
			  gboolean *can_write,
			  gboolean *is_blank,
			  gboolean *is_DVD)
{
	NautilusBurnMediaType real_type;
	gboolean is_blank_real = FALSE;

	/* if drive is mounted then unmount before checking anything */
	if (nautilus_burn_drive_is_mounted (drive)) {
		if (!nautilus_burn_drive_unmount (drive))
			g_warning ("Couldn't unmount volume in drive: %s", drive->device);
	}

	real_type = nautilus_burn_drive_get_media_type_full (drive,
							     is_rewritable,
							     &is_blank_real,
							     NULL,
							     NULL);
	if (type)
		*type = real_type;
	if (is_blank)
		*is_blank = is_blank_real;
	if (can_write)
		*can_write = nautilus_burn_drive_media_type_is_writable (real_type,
									 is_blank_real);

       if (is_DVD) {
		if (real_type > NAUTILUS_BURN_MEDIA_TYPE_CDRW)
			*is_DVD = TRUE;
		else
			*is_DVD = FALSE;
	}
}

static BonfireBurnResult
bonfire_burn_ask_for_media (BonfireBurn *burn,
			    NautilusBurnDrive *drive,
			    BonfireBurnError error_type,
			    BonfireMediaType required_media,
			    gboolean eject,
			    GError **error)
{
	gint64 media_size;
	gboolean is_reload;
	gboolean is_mounted;
	BonfireBurnResult result;
	NautilusBurnMediaType type;

	media_size = nautilus_burn_drive_get_media_size (drive);
	type = nautilus_burn_drive_get_media_type (drive);

	is_reload = (media_size > 0);
	if (type == NAUTILUS_BURN_MEDIA_TYPE_ERROR)
		is_reload = FALSE;

	/* check one more time */
	is_mounted = nautilus_burn_drive_is_mounted (drive);

	if (is_mounted == TRUE)
		error_type = BONFIRE_BURN_ERROR_MEDIA_BUSY;
	if (is_reload == FALSE)
		error_type = BONFIRE_BURN_ERROR_MEDIA_NONE;

	if (type != NAUTILUS_BURN_MEDIA_TYPE_ERROR
	&&  GPOINTER_TO_INT (g_object_get_data (G_OBJECT (drive), IS_LOCKED))
	&&  !nautilus_burn_drive_unlock (drive)) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     _("the drive can't be unlocked"));
		return BONFIRE_BURN_ERROR;
	}

	g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (0));

	if (eject && !nautilus_burn_drive_eject (drive)) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     _("the disc can't be ejected"));
		return BONFIRE_BURN_ERR;
	}

	result = BONFIRE_BURN_CANCEL;
	g_signal_emit (burn,
		       bonfire_burn_signals [INSERT_MEDIA_REQUEST_SIGNAL],
		       0,
		       error_type,
		       required_media,
		       &result);

	return result;
}

BonfireBurnResult
bonfire_burn_wait_for_source_media (BonfireBurn *burn,
				    NautilusBurnDrive *drive,
				    gboolean *is_DVD,
				    gboolean eject,
				    GError **error)
{
	char *failure;
	gboolean is_blank;
	BonfireBurnResult result;
	NautilusBurnMediaType type;

 again:

	bonfire_burn_check_media (burn,
				  drive,
				  &type,
				  NULL,
				  NULL,
				  &is_blank,
				  is_DVD);

	if (is_blank) {
		result = bonfire_burn_ask_for_media (burn,
						     drive,
						     BONFIRE_BURN_ERROR_MEDIA_BLANK,
						     BONFIRE_MEDIA_WITH_DATA,
						     eject,
						     error);
		if (result != BONFIRE_BURN_OK)
			return result;

		goto again;
	}

	if (!GPOINTER_TO_INT (g_object_get_data (G_OBJECT (drive), IS_LOCKED))
	&&  !nautilus_burn_drive_lock (drive, _("ongoing copying process"), &failure)) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     failure);
		return BONFIRE_BURN_ERR;
	}
	g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (1));

	return BONFIRE_BURN_OK;
}

static BonfireBurnResult
bonfire_burn_wait_for_rewritable_media (BonfireBurn *burn,
					NautilusBurnDrive *drive,
					gboolean *is_DVD,
					gboolean eject,
					GError **error)
{
	char *failure;
	gboolean is_blank;
	gboolean can_write;
	gboolean is_rewritable;
	BonfireBurnResult result;
	NautilusBurnMediaType type;

	if (!nautilus_burn_drive_can_rewrite (drive)) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     _("the drive has no rewriting capabilities"));
		return BONFIRE_BURN_NOT_SUPPORTED;
	}

 again:

	bonfire_burn_check_media (burn,
				  drive,
				  &type,
				  &is_rewritable,
				  &can_write,
				  &is_blank,
				  is_DVD);

	if (is_blank || !is_rewritable) {
		result = bonfire_burn_ask_for_media (burn,
						     drive,
						     is_blank ? BONFIRE_BURN_ERROR_MEDIA_BLANK : BONFIRE_BURN_ERROR_MEDIA_NOT_REWRITABLE,
						     BONFIRE_MEDIA_REWRITABLE|
						     BONFIRE_MEDIA_WITH_DATA,
						     eject,
						     error);
		if (result != BONFIRE_BURN_OK)
			return result;

		goto again;
	}

	if (!GPOINTER_TO_INT (g_object_get_data (G_OBJECT (drive), IS_LOCKED))
	&&  !nautilus_burn_drive_lock (drive, _("ongoing blanking process"), &failure)) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     failure);
		return BONFIRE_BURN_ERR;
	}

	g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (1));

	return BONFIRE_BURN_OK;
}

static BonfireBurnResult
bonfire_burn_blank_real (BonfireBurn *burn,
			 NautilusBurnDrive *drive,
			 BonfireRecorderFlag flags,
			 gboolean is_DVD,
			 gboolean is_debug,
			 GError **error)
{
	BonfireBurnResult result;

	if (is_DVD) {
		NautilusBurnMediaType type;

		type = nautilus_burn_drive_get_media_type (drive);

		if (type == NAUTILUS_BURN_MEDIA_TYPE_DVDRW
		||  type == NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW)
			burn->priv->recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_GROWISOFS, NULL));
		else
			burn->priv->recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_DVD_RW_FORMAT, NULL));
	}
	else
		burn->priv->recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_CD_RECORD, NULL));

	g_signal_connect (burn->priv->recorder,
			  "progress_changed",
			  G_CALLBACK (bonfire_burn_progress_changed),
			  burn);
	g_signal_connect (burn->priv->recorder,
			  "action_changed",
			  G_CALLBACK (bonfire_burn_action_changed),
			  burn);

	bonfire_job_set_debug (BONFIRE_JOB (burn->priv->recorder), is_debug);
	result = bonfire_recorder_set_drive (burn->priv->recorder,
					     drive,
					     error);
	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_recorder_set_flags (burn->priv->recorder,
					     flags,
					     error);
	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_recorder_blank (burn->priv->recorder, error);
	g_object_unref (burn->priv->recorder);
	burn->priv->recorder = NULL;

	return result;
}

BonfireBurnResult
bonfire_burn_blank (BonfireBurn *burn,
		    BonfireBurnFlag burn_flags,
		    NautilusBurnDrive *drive,
		    gboolean fast,
		    GError **error)
{
	BonfireRecorderFlag blank_flags;
	BonfireBurnResult result;
	GError *ret_error = NULL;
	gboolean is_DVD = FALSE;

	/* we wait for the insertion of a media and lock it */
	g_return_val_if_fail (drive != NULL, BONFIRE_BURN_ERR);
	result = bonfire_burn_wait_for_rewritable_media (burn,
							 drive,
							 &is_DVD,
							 (burn_flags & BONFIRE_BURN_FLAG_EJECT),
							 error);
	if (result != BONFIRE_BURN_OK)
		return result;

	blank_flags = BONFIRE_RECORDER_FLAG_NONE;
	if (burn_flags & BONFIRE_BURN_FLAG_NOGRACE)
		blank_flags |= BONFIRE_RECORDER_FLAG_NOGRACE;

	if (burn_flags & BONFIRE_BURN_FLAG_DUMMY)
		blank_flags |= BONFIRE_RECORDER_FLAG_DUMMY;

	if (fast)
		blank_flags |= BONFIRE_RECORDER_FLAG_FAST_BLANK;

	result = bonfire_burn_blank_real (burn,
					  drive,
					  blank_flags,
					  is_DVD,
					  (burn_flags & BONFIRE_BURN_FLAG_DEBUG) != 0,
					  &ret_error);

	while (result == BONFIRE_BURN_ERR
	&&     ret_error
	&&     ret_error->code == BONFIRE_BURN_ERROR_MEDIA_NOT_REWRITABLE) {
		g_error_free (ret_error);
		ret_error = NULL;
		
		result = bonfire_burn_ask_for_media (burn,
						     drive,
						     BONFIRE_BURN_ERROR_MEDIA_NOT_REWRITABLE,
						     BONFIRE_MEDIA_REWRITABLE|
						     BONFIRE_MEDIA_WITH_DATA,
						     (burn_flags & BONFIRE_BURN_FLAG_EJECT),
						     error);
		if (result != BONFIRE_BURN_OK)
			break;

		result = bonfire_burn_wait_for_rewritable_media (burn,
								 drive,
								 &is_DVD,
								 (burn_flags & BONFIRE_BURN_FLAG_EJECT),
								 error);
		if (result != BONFIRE_BURN_OK)
			return result;

		result = bonfire_burn_blank_real (burn,
						  drive,
						  blank_flags,
						  is_DVD,
						  (burn_flags & BONFIRE_BURN_FLAG_DEBUG) != 0,
						  &ret_error);
	}

	if (ret_error)
		g_propagate_error (error, ret_error);

	nautilus_burn_drive_unlock (drive);
	g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (0));

	if (burn_flags & BONFIRE_BURN_FLAG_EJECT)
		bonfire_burn_common_eject_async (drive);

	if (result == BONFIRE_BURN_OK)
		bonfire_burn_action_changed_real (burn,
						  BONFIRE_BURN_ACTION_FINISHED,
						  NULL);
	return result;
}

static BonfireBurnResult
bonfire_burn_wait_for_recordable_media (BonfireBurn *burn,
					BonfireBurnFlag flags,
					NautilusBurnDrive *drive,
					BonfireTrackSourceType track_type,
					gboolean *is_DVD,
					gboolean *is_rewritable,
					GError **error)
{
	char *failure;
	gboolean is_blank;
	gboolean can_write;
	gboolean is_DVD_real;
	BonfireBurnResult result;
	NautilusBurnMediaType type;
	gboolean is_rewritable_real;
	BonfireBurnError berror;

	if (!nautilus_burn_drive_can_write (drive)) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     _("the drive has no burning capabilities"));
		return BONFIRE_BURN_NOT_SUPPORTED;
	}

	result = BONFIRE_BURN_OK;

 again:

	berror = BONFIRE_BURN_ERROR_NONE;
	bonfire_burn_check_media (burn,
				  drive,
				  &type,
				  &is_rewritable_real,
				  &can_write,
				  &is_blank,
				  &is_DVD_real);

	if (GPOINTER_TO_INT (g_object_get_data (G_OBJECT (drive), IS_LOCKED))) {
		if (is_DVD)
			*is_DVD = is_DVD_real;

		if (is_rewritable)
			*is_rewritable = is_rewritable_real;

		/* the drive has already been checked just return its characteristics */
		/* NOTE: after a blanking, for nautilus_burn the CD/DVD is still full of
		 * data so if the drive has already been checked there is no need to do
		 * that again since we would be asked if we want to blank it again */
		return result;
	}

	if (type == NAUTILUS_BURN_MEDIA_TYPE_BUSY) {
		result = BONFIRE_BURN_NEED_RELOAD;
		berror = BONFIRE_BURN_ERROR_MEDIA_BUSY;
		goto end;
	}

	if (!can_write) {
		result = BONFIRE_BURN_NEED_RELOAD;
		berror = BONFIRE_BURN_ERROR_MEDIA_NOT_WRITABLE;
		goto end;
	}

	if ((track_type != BONFIRE_TRACK_SOURCE_ISO
	&&   track_type != BONFIRE_TRACK_SOURCE_ISO_JOLIET
	&&   track_type != BONFIRE_TRACK_SOURCE_DATA
	&&   track_type != BONFIRE_TRACK_SOURCE_DISC)
	&&  type > NAUTILUS_BURN_MEDIA_TYPE_CDRW) {
		result = BONFIRE_BURN_NEED_RELOAD;
		berror = BONFIRE_BURN_ERROR_DVD_NOT_SUPPORTED;
		goto end;	
	}

	/* FIXME: we should check that the image will fit on the media */
	if (!nautilus_burn_drive_lock (drive, _("ongoing burning process"), &failure)) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     _("the drive can't be locked (%s)"),
			     failure);
		return BONFIRE_BURN_ERR;
	}

	g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (1));

	if (!nautilus_burn_drive_can_rewrite (drive))
		is_rewritable_real = FALSE;

	if (is_DVD)
		*is_DVD = is_DVD_real;

	/* silently ignore if the drive is not rewritable */
	/* NOTE: we corrected such contradictory flags setting (might be an error) as
	 * BONFIRE_BURN_FLAG_MERGE|BONFIRE_BURN_FLAG_APPEND|BONFIRE_BURN_FLAG_BLANK_BEFORE_WRITE */
	/* FIXME: strange according to DVD_RW are blank ?? */
	if ((flags & BONFIRE_BURN_FLAG_BLANK_BEFORE_WRITE)
	&&  is_rewritable_real
	&&  !is_blank) {
		g_signal_emit (burn,
			       bonfire_burn_signals [WARN_DATA_LOSS_SIGNAL],
			       0,
			       &result);

		if (result != BONFIRE_BURN_OK)
			goto end;
	
		if (type != NAUTILUS_BURN_MEDIA_TYPE_DVD_PLUS_RW) {
			BonfireRecorderFlag blank_flags = BONFIRE_RECORDER_FLAG_FAST_BLANK;

			if (flags & BONFIRE_BURN_FLAG_DUMMY)
				blank_flags |= BONFIRE_RECORDER_FLAG_DUMMY;
	
			if (flags & BONFIRE_BURN_FLAG_NOGRACE)
				blank_flags |= BONFIRE_RECORDER_FLAG_NOGRACE;
	
			result = bonfire_burn_blank_real (burn,
							  drive,
							  blank_flags,
							  is_DVD_real,
							  (flags & BONFIRE_BURN_FLAG_DEBUG) != 0,
							  error);
		}
	}

	if (is_rewritable)
		*is_rewritable = is_rewritable_real;

end:

	if (result == BONFIRE_BURN_NEED_RELOAD) {
		BonfireMediaType required_media;

		if (track_type != BONFIRE_TRACK_SOURCE_ISO
		||  track_type != BONFIRE_TRACK_SOURCE_ISO_JOLIET
		||  track_type != BONFIRE_TRACK_SOURCE_DATA
		||  track_type != BONFIRE_TRACK_SOURCE_DISC)
			required_media = BONFIRE_MEDIA_WRITABLE|
					 BONFIRE_MEDIA_TYPE_CD;
		else
			required_media = BONFIRE_MEDIA_WRITABLE;

		result = bonfire_burn_ask_for_media (burn,
						     drive,
						     berror,
						     required_media,
						     (flags & BONFIRE_BURN_FLAG_EJECT),
						     error);
		if (result == BONFIRE_BURN_OK)
			goto again;
	}

	if (result != BONFIRE_BURN_OK) {
		g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (0));
		nautilus_burn_drive_unlock (drive);
	}

	return result;
}

static BonfireBurnResult
bonfire_burn_create_recorder (BonfireBurn *burn,
			      BonfireBurnFlag flags,
			      NautilusBurnDrive *drive,
			      BonfireTrackSource *track,
			      gint speed,
			      gboolean is_rewritable,
			      gboolean is_DVD,
			      GError **error)
{
	BonfireBurnResult result;
	BonfireRecorderFlag rec_flags;
	BonfireRecorder *recorder = NULL;

	switch (track->type) {
	case BONFIRE_TRACK_SOURCE_IMAGER:
		if (BONFIRE_IS_READCD (track->contents.imager.obj)) {
			if (is_DVD)
				recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_GROWISOFS, NULL));
			else
				recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_CD_RECORD, NULL));
		}
		else if (BONFIRE_IS_TRANSCODE (track->contents.imager.obj)) {
			if (is_rewritable) {
				g_signal_emit (burn,
					       bonfire_burn_signals [WARN_REWRITABLE_SIGNAL],
					       0,
					       &result);

				if (result != BONFIRE_BURN_OK)
					return result;
			}

			recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_CD_RECORD, NULL));
		}
		else if (BONFIRE_IS_GROWISOFS (track->contents.imager.obj)) {
			recorder = BONFIRE_RECORDER (track->contents.imager.obj);
			g_object_ref (recorder);
			track = NULL;
		}
		else if (BONFIRE_IS_CDRDAO (track->contents.imager.obj)) {
			recorder = BONFIRE_RECORDER (track->contents.imager.obj);
			g_object_ref (recorder);
			track = NULL;
		}
		else if (BONFIRE_IS_MKISOFS (track->contents.imager.obj))
			recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_CD_RECORD, NULL));

		break;

	case BONFIRE_TRACK_SOURCE_AUDIO:
		if (is_DVD)
			return BONFIRE_BURN_NOT_SUPPORTED;

		if (is_rewritable) {
			g_signal_emit (burn,
				       bonfire_burn_signals [WARN_REWRITABLE_SIGNAL],
				       0,
				       &result);

			if (result != BONFIRE_BURN_OK)
				return result;
		}

		recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_CD_RECORD, NULL));
		break;

	case BONFIRE_TRACK_SOURCE_ISO:
	case BONFIRE_TRACK_SOURCE_ISO_JOLIET:
		if (is_DVD)
			recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_GROWISOFS, NULL));
		else
			recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_CD_RECORD, NULL));
		break;

	case BONFIRE_TRACK_SOURCE_CUE:
		if (is_DVD)
			return BONFIRE_BURN_NOT_SUPPORTED;

		recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_CDRDAO, NULL));
		break;

	case BONFIRE_TRACK_SOURCE_RAW:
		if (is_DVD)
			return BONFIRE_BURN_NOT_SUPPORTED;

		recorder = BONFIRE_RECORDER (g_object_new (BONFIRE_TYPE_CD_RECORD, NULL));
		break;

	default:
		return BONFIRE_BURN_NOT_SUPPORTED;
	}

	if (!recorder)
		return BONFIRE_BURN_ERR;

	if ((flags & BONFIRE_BURN_FLAG_DEBUG) != 0)
		bonfire_job_set_debug (BONFIRE_JOB (recorder), TRUE);

	burn->priv->recorder = recorder;
	rec_flags = BONFIRE_RECORDER_FLAG_NONE;

	if (flags & BONFIRE_BURN_FLAG_NOGRACE)
		rec_flags |= BONFIRE_RECORDER_FLAG_NOGRACE;

	if (flags & BONFIRE_BURN_FLAG_OVERBURN)
		rec_flags |= BONFIRE_RECORDER_FLAG_OVERBURN;

	if (flags & BONFIRE_BURN_FLAG_BURNPROOF)
		rec_flags |= BONFIRE_RECORDER_FLAG_BURNPROOF;

	if (flags & BONFIRE_BURN_FLAG_DAO)
		rec_flags |= BONFIRE_RECORDER_FLAG_DAO;

	if (flags & BONFIRE_BURN_FLAG_DONT_CLOSE)
		rec_flags |= BONFIRE_RECORDER_FLAG_MULTI;

	if (flags & BONFIRE_BURN_FLAG_DUMMY)
		rec_flags |= BONFIRE_RECORDER_FLAG_DUMMY;

	result = bonfire_job_set_rate (BONFIRE_JOB (recorder),
				       is_DVD ? speed * DVD_SPEED : speed * CDR_SPEED);
	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_recorder_set_drive (recorder,
					     drive,
					     error);
	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_recorder_set_flags (recorder,
					     rec_flags,
					     error);
	if (result != BONFIRE_BURN_OK)
		return result;

	if (track)
		result = bonfire_recorder_set_track (recorder,
						     track,
						     error);

	g_signal_connect (burn->priv->recorder,
			  "progress_changed",
			  G_CALLBACK (bonfire_burn_progress_changed),
			  burn);
	g_signal_connect (burn->priv->recorder,
			  "action_changed",
			  G_CALLBACK (bonfire_burn_action_changed),
			  burn);
	return result;
}

static BonfireBurnResult
bonfire_burn_reload_record_media (BonfireBurn *burn,
				  BonfireBurnError error_code,
				  BonfireBurnFlag flags,
				  NautilusBurnDrive *drive,
				  BonfireTrackSource *track,
				  gint speed,
				  gboolean eject,
				  GError **error)
{
	BonfireMediaType required_media;
	BonfireBurnResult result;
	gboolean is_rewritable;
	gboolean is_DVD;

again:
	if (burn->priv->recorder) {
		g_object_unref (burn->priv->recorder);
		burn->priv->recorder = NULL;
	}

	if (track->type != BONFIRE_TRACK_SOURCE_ISO
	||  track->type != BONFIRE_TRACK_SOURCE_ISO_JOLIET
	||  track->type != BONFIRE_TRACK_SOURCE_DATA
	||  track->type != BONFIRE_TRACK_SOURCE_DISC)
		required_media = BONFIRE_MEDIA_WRITABLE|
				 BONFIRE_MEDIA_TYPE_CD;
	else
		required_media = BONFIRE_MEDIA_WRITABLE;

	result = bonfire_burn_ask_for_media (burn,
					     drive,
					     error_code,
					     required_media,
					     (flags & BONFIRE_BURN_FLAG_EJECT),
					     error);
	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_burn_wait_for_recordable_media (burn,
							 flags,
							 drive,
							 track->type,
							 &is_DVD,
							 &is_rewritable,
							 error);
	if (result == BONFIRE_BURN_NEED_RELOAD)
		goto again;

	if (result != BONFIRE_BURN_OK)
		return result;

	/* the media might have changed CD->DVD or DVD->CD */
	result = bonfire_burn_create_recorder (burn,
					       flags,
					       drive,
					       track,
					       speed,
					       is_rewritable,
					       is_DVD,
					       error);

	if (result == BONFIRE_BURN_NEED_RELOAD)
		goto again;

	return result;
}

static BonfireBurnResult
bonfire_burn_setup_imager (BonfireBurn *burn,
			   BonfireBurnFlag flags,
			   NautilusBurnDrive *drive,
			   const BonfireTrackSource *source,
			   BonfireTrackSourceType track_type,
			   const char *output,
			   GError **error)
{
	BonfireBurnResult result;

	if (!burn->priv->imager)
		return BONFIRE_BURN_OK;

	g_signal_connect (burn->priv->imager,
			  "progress_changed",
			  G_CALLBACK (bonfire_burn_progress_changed),
			  burn);
	g_signal_connect (burn->priv->imager,
			  "action_changed",
			  G_CALLBACK (bonfire_burn_action_changed),
			  burn);

	result = bonfire_imager_set_source (burn->priv->imager,
					    source,
					    error);
	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_imager_set_output_type (burn->priv->imager,
						 track_type,
						 error);
	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_imager_set_output (burn->priv->imager,
					    output,
					    (flags & BONFIRE_BURN_FLAG_DONT_OVERWRITE) == 0,
					    (flags & BONFIRE_BURN_FLAG_DONT_CLEAN_OUTPUT) == 0,
					    error);

	if ((result != BONFIRE_BURN_OK && result != BONFIRE_BURN_NOT_SUPPORTED)
	||  (result == BONFIRE_BURN_NOT_SUPPORTED && output != NULL))
		return result;

	if ((flags & (BONFIRE_BURN_FLAG_MERGE|BONFIRE_BURN_FLAG_APPEND))) {
		result = bonfire_imager_set_append (burn->priv->imager,
						    drive,
						    (flags & BONFIRE_BURN_FLAG_MERGE) != 0,
						    error);
		if (result != BONFIRE_BURN_OK)
			return result;
	}

	return BONFIRE_BURN_OK;
}

static BonfireBurnResult
bonfire_burn_create_imager (BonfireBurn *burn,
			    BonfireBurnFlag flags,
			    NautilusBurnDrive *drive,
			    const BonfireTrackSource *source,
			    BonfireTrackSourceType *track_type,
			    GError **error)
{
	BonfireBurnResult result = BONFIRE_BURN_OK;
	BonfireTrackSourceType type;
	gboolean is_DVD;

	switch (source->type) {
	case BONFIRE_TRACK_SOURCE_SONG:
		type = BONFIRE_TRACK_SOURCE_AUDIO;

		/* FIXME: maybe we should do as below if it's on_the_fly */
		burn->priv->imager = BONFIRE_IMAGER (g_object_new (BONFIRE_TYPE_TRANSCODE, NULL));
		break;

	case BONFIRE_TRACK_SOURCE_GRAFTS:
	case BONFIRE_TRACK_SOURCE_DATA:
		type = BONFIRE_TRACK_SOURCE_ISO_JOLIET;
		is_DVD = FALSE;

		/* if we append/merge/burn on the fly we need to make sure now
		 * there is a disc in the drive so as to lock it, because:
		 * - if it's a DVD we'll use 
		 * - to append or merge cdr* programs need to know where to start the track */
		if (flags & (BONFIRE_BURN_FLAG_ON_THE_FLY|BONFIRE_BURN_FLAG_MERGE|BONFIRE_BURN_FLAG_APPEND)) {
			result = bonfire_burn_wait_for_recordable_media (burn,
									 flags,
									 drive,
									 source->type,
									 &is_DVD,
									 NULL,
									 error);
			if (result != BONFIRE_BURN_OK)
				return result;
		}
	
		if (is_DVD) {
			burn->priv->imager = BONFIRE_IMAGER (g_object_new (BONFIRE_TYPE_GROWISOFS, NULL));

			result = bonfire_recorder_set_drive (BONFIRE_RECORDER (burn->priv->imager), drive, error);
			if (result != BONFIRE_BURN_OK)
				return result;
		}
		else
			burn->priv->imager = BONFIRE_IMAGER (g_object_new (BONFIRE_TYPE_MKISOFS, NULL));
		break;

	case BONFIRE_TRACK_SOURCE_DISC:
		result = bonfire_burn_wait_for_source_media (burn,
							     source->contents.drive.disc,
							     &is_DVD,
							     (flags & BONFIRE_BURN_FLAG_EJECT),
							     error);
		if (result != BONFIRE_BURN_OK)
			return result;

		/* NOTE: is_DVD refers to the media we should copy here not to the recordable disc */
		if (is_DVD) {
			burn->priv->imager = BONFIRE_IMAGER (g_object_new (BONFIRE_TYPE_READCD, NULL));
			type = BONFIRE_TRACK_SOURCE_ISO;
		}
		else {
			burn->priv->imager = BONFIRE_IMAGER (g_object_new (BONFIRE_TYPE_CDRDAO, NULL));
			type = BONFIRE_TRACK_SOURCE_CUE;
		}

		break;

	case BONFIRE_TRACK_SOURCE_INF: /* no need for an imager object */
	case BONFIRE_TRACK_SOURCE_CUE:
	case BONFIRE_TRACK_SOURCE_RAW:
	case BONFIRE_TRACK_SOURCE_ISO:
	case BONFIRE_TRACK_SOURCE_ISO_JOLIET:
		*track_type = source->type;
		return BONFIRE_BURN_OK;

	default:
		return BONFIRE_BURN_NOT_SUPPORTED;
	}

	if ((flags & BONFIRE_BURN_FLAG_DEBUG) != 0)
		bonfire_job_set_debug (BONFIRE_JOB (burn->priv->imager), TRUE);

	if (result == BONFIRE_BURN_OK)
		*track_type = type;
	else
		*track_type = BONFIRE_TRACK_SOURCE_UNKNOWN;

	return BONFIRE_BURN_OK;
}
static BonfireBurnResult
bonfire_burn_ask_for_joliet (BonfireBurn *burn)
{
	BonfireBurnResult result;

	g_signal_emit (burn,
		       bonfire_burn_signals [ASK_DISABLE_JOLIET_SIGNAL],
		       0,
		       &result);

	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_imager_set_output_type (burn->priv->imager,
						 BONFIRE_TRACK_SOURCE_ISO,
						 NULL);

	return result;
}

static BonfireBurnResult
bonfire_burn_check_volume_free_space (BonfireBurn *burn,
				      const char *output,
				      GError **error)
{
	char *dirname;
	char *uri_str;
	GnomeVFSURI *uri;
	GnomeVFSResult result;
	GnomeVFSFileSize vol_size;

	if (!output)
		dirname = g_strdup (g_get_tmp_dir ());
	else
		dirname = g_path_get_dirname (output);

	uri_str = gnome_vfs_get_uri_from_local_path (dirname);
	g_free (dirname);

	uri = gnome_vfs_uri_new (uri_str);
	g_free (uri_str);

	if (uri == NULL)
		return BONFIRE_BURN_ERR;

	result = gnome_vfs_get_volume_free_space (uri, &vol_size);
	if (result != GNOME_VFS_OK) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     _("the size of the volume can't be checked (%s)"),
			     gnome_vfs_result_to_string (result));
		gnome_vfs_uri_unref (uri);
		return BONFIRE_BURN_ERR;
	}
	gnome_vfs_uri_unref (uri);

	if (burn->priv->image_size > vol_size) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_DISC_SPACE,
			     _("the selected location does not have enough free space to store the disc image (%ld MiB needed)"),
			     (unsigned long) burn->priv->image_size / 1048576);
		return BONFIRE_BURN_ERR;
	}

	return BONFIRE_BURN_OK;
}

static BonfireBurnResult
bonfire_burn_imager_get_track (BonfireBurn *burn,
			       BonfireBurnFlag flags,
			       NautilusBurnDrive *drive,
			       const BonfireTrackSource *source,
			       BonfireTrackSource **track,
			       const char *output,
			       GError **error)
{
	BonfireTrackSourceType track_type;
	BonfireBurnError error_code;
	BonfireBurnResult result;
	GError *ret_error = NULL;

	result = bonfire_burn_create_imager (burn,
					     flags,
					     drive,
					     source,
					     &track_type,
					     error);
	if (result != BONFIRE_BURN_OK)
		return result;

	result = bonfire_burn_setup_imager (burn,
					    flags,
					    drive,
					    source,
					    track_type,
					    output,
					    error);
	if (result != BONFIRE_BURN_OK)
		return result;

 	/* we get the size and (will) check that if we're writing to a disc,
	 * the media is big enough and/or if we're writing to a hard drive,
	 * the volume is big enough */
	result = bonfire_imager_get_size (BONFIRE_IMAGER (burn->priv->imager),
					  &burn->priv->image_size,
					  FALSE,
					  &ret_error);
	if (result == BONFIRE_BURN_ERR) {
		error_code = ret_error->code;

		if (error_code == BONFIRE_BURN_ERROR_JOLIET_TREE)
			result = bonfire_burn_ask_for_joliet (burn);

		if (result == BONFIRE_BURN_OK) {
			g_error_free (ret_error);
			ret_error = NULL;

			result = bonfire_imager_get_size (BONFIRE_IMAGER (burn->priv->imager),
							  &burn->priv->image_size,
							  FALSE,
							  &ret_error);
		}
		else
			g_propagate_error (error, ret_error);
	}

	if (result != BONFIRE_BURN_OK)
		return result;

	if (flags & BONFIRE_BURN_FLAG_ON_THE_FLY
	||  BONFIRE_IS_GROWISOFS (burn->priv->imager)) {
		BonfireTrackSource *source;

		source = g_new0 (BonfireTrackSource, 1);
		source->type = BONFIRE_TRACK_SOURCE_IMAGER;
		source->contents.imager.obj = burn->priv->imager;
		burn->priv->imager = NULL;

		*track = source;
		return BONFIRE_BURN_OK;
	}

	if (flags & BONFIRE_BURN_FLAG_CHECK_SIZE) {
		result = bonfire_burn_check_volume_free_space (burn,
							       output,
							       error);
	
		if (result != BONFIRE_BURN_OK)
			return result;
	}

	while (1) {
		result = bonfire_imager_get_track (burn->priv->imager,
						   track,
						   &ret_error);
		if (result == BONFIRE_BURN_OK)
			break;

		if (result != BONFIRE_BURN_ERR)
			break;

		/* See if error is recoverable */
		error_code = ret_error->code;
		if (error_code == BONFIRE_BURN_ERROR_JOLIET_TREE) {
			if (bonfire_burn_ask_for_joliet (burn) != BONFIRE_BURN_OK) {
				g_propagate_error (error, ret_error);
				break;
			}

			continue;
		}

		if (error_code != BONFIRE_BURN_ERROR_MEDIA_BLANK) {
			g_propagate_error (error, ret_error);
			break;
		}

		g_error_free (ret_error);
		ret_error = NULL;

		/* we recreate an imager in case it is not the same type of disc,
		 * that's why we also recheck the size */
		g_object_unref (burn->priv->imager);
		burn->priv->imager = NULL;

		result = bonfire_imager_get_size (BONFIRE_IMAGER (burn->priv->imager),
						  &burn->priv->image_size,
						  FALSE,
						  error);
		if (result != BONFIRE_BURN_OK)
			break;

		if (flags & BONFIRE_BURN_FLAG_CHECK_SIZE) {
			result = bonfire_burn_check_volume_free_space (burn,
								       output,
								       error);
		
			if (result != BONFIRE_BURN_OK)
				break;
		}

		result = bonfire_burn_ask_for_media (burn,
						     source->contents.drive.disc,
						     error_code,
						     BONFIRE_MEDIA_WITH_DATA,
						     (flags & BONFIRE_BURN_FLAG_EJECT),
						     error);
		if (result != BONFIRE_BURN_OK)
			break;

		result = bonfire_burn_create_imager (burn,
						     flags,
						     drive,
						     source,
						     &track_type,
						     error);
		if (result != BONFIRE_BURN_OK)
			break;

		result = bonfire_burn_setup_imager (burn,
						    flags,
						    drive,
						    source,
						    track_type,
						    output,
						    error);
		if (result != BONFIRE_BURN_OK)
			break;
	}

	return result;
}

static BonfireBurnResult
bonfire_burn_get_size (BonfireBurn *burn,
		       BonfireBurnFlag flags,
		       const BonfireTrackSource *source,
		       GError **error)
{
	BonfireBurnResult result = BONFIRE_BURN_OK;
	GnomeVFSFileInfo *info;
	GnomeVFSResult res;
	char *uri = NULL;

	switch (source->type) {
	case BONFIRE_TRACK_SOURCE_INF:
		return BONFIRE_BURN_NOT_SUPPORTED;

	case BONFIRE_TRACK_SOURCE_CUE:
		uri = g_strconcat ("file://", source->contents.cue.image, NULL);
		break;

	case BONFIRE_TRACK_SOURCE_RAW:
		uri = g_strconcat ("file://", source->contents.raw.image, NULL);
		break;

	case BONFIRE_TRACK_SOURCE_ISO:
	case BONFIRE_TRACK_SOURCE_ISO_JOLIET:
		uri = g_strconcat ("file://", source->contents.iso.image, NULL);
		break;

	default:
		return BONFIRE_BURN_NOT_SUPPORTED;	
	}

	info = gnome_vfs_file_info_new ();
	res = gnome_vfs_get_file_info (uri, 
				       info,
				       GNOME_VFS_FILE_INFO_DEFAULT);

	burn->priv->image_size = info->size;

	gnome_vfs_file_info_clear (info);
	gnome_vfs_file_info_unref (info);
	g_free (uri);

	if (res != GNOME_VFS_OK) {
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     _("the file %s can't be opened (%s)"),
			     uri,
			     gnome_vfs_result_to_string (res));
		burn->priv->image_size = -1;
		return BONFIRE_BURN_ERR;
	}

	return result;
}

/* FIXME: for the moment we don't allow for mixed CD type */
/* FIXME: for the moment we can't deal with non local images (cues, iso, bin) maybe a new object ImagerNull */
BonfireBurnResult 
bonfire_burn_record (BonfireBurn *burn,
		     BonfireBurnFlag flags,
		     NautilusBurnDrive *drive,
		     gint speed,
		     const BonfireTrackSource *source,
		     const char *output,
		     GError **error)
{
	gboolean same_drive;
	BonfireBurnResult result;
	BonfireTrackSource *track = NULL;

	g_return_val_if_fail (drive != NULL, BONFIRE_BURN_ERR);

	same_drive = FALSE;
	
	/* we check flags consistency 
	 * NOTE: should we return an error if they are not consistent? */
	if ((source->type != BONFIRE_TRACK_SOURCE_SONG
	&&   source->type != BONFIRE_TRACK_SOURCE_DATA
	&&   source->type != BONFIRE_TRACK_SOURCE_GRAFTS
	&&   source->type != BONFIRE_TRACK_SOURCE_DISC)
	||  drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE) {
		if (flags & BONFIRE_BURN_FLAG_MERGE) {
			g_warning ("Inconsistent flag: you can't use flag merge\n");
			flags &= ~BONFIRE_BURN_FLAG_MERGE;
		}
			
		if (flags & BONFIRE_BURN_FLAG_APPEND) {
			g_warning ("Inconsistent flags: you can't use flag append\n");
			flags &= ~BONFIRE_BURN_FLAG_APPEND;
		}

		if (flags & BONFIRE_BURN_FLAG_ON_THE_FLY) {
			g_warning ("Inconsistent flag: you can't use flag on_the_fly\n");
			flags &= ~BONFIRE_BURN_FLAG_ON_THE_FLY;
		}
	}

	if ((flags & (BONFIRE_BURN_FLAG_MERGE|BONFIRE_BURN_FLAG_APPEND)) != 0
	&&  (flags & BONFIRE_BURN_FLAG_BLANK_BEFORE_WRITE) != 0) {
		g_warning ("Inconsistent flag: you can't use flag blank_before_write\n");
		flags &= ~BONFIRE_BURN_FLAG_BLANK_BEFORE_WRITE;
	}

	if (drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE
	&& (flags & BONFIRE_BURN_FLAG_DONT_CLEAN_OUTPUT) == 0) {
		g_warning ("Inconsistent flag: you can't use flag blank_before_write\n");
		flags |= BONFIRE_BURN_FLAG_DONT_CLEAN_OUTPUT;
	}

	/* this is a special case when source is the same drive as the recording drive */
	if (source->type == BONFIRE_TRACK_SOURCE_DISC
	&&  nautilus_burn_drive_equal (drive, source->contents.drive.disc)) {
		same_drive = TRUE;
	}

	nautilus_burn_drive_ref (drive);
	burn->priv->drive = drive;

	/* create needed imager for certain type of tracks (grafts, audio, disc) */
	if (source->type == BONFIRE_TRACK_SOURCE_DATA
	||  source->type == BONFIRE_TRACK_SOURCE_GRAFTS
	||  source->type == BONFIRE_TRACK_SOURCE_SONG
	||  source->type == BONFIRE_TRACK_SOURCE_DISC) {
		result = bonfire_burn_imager_get_track (burn,
							flags,
							drive,
							source,
							&track,
							output,
							error);
	
		if (result != BONFIRE_BURN_OK)
			goto end;
	}
	else {
		/* we get the size and (will) check that if we're writing to a disc,
		 * the media is big enough and/or if we're writing to a hard drive,
		 * the volume is big enough */
		result = bonfire_burn_get_size (burn,
						flags,
						source,
						error);
		if (result != BONFIRE_BURN_OK)
			goto end;

		track = bonfire_track_source_copy (source);
	}

	if (drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE)
		goto end;

	/* now record */
	if (same_drive) {
		/* NOTE: we use track->contents.drive.disc here so as to keep
		 * the IS_LOCKED value consistent */
		if (GPOINTER_TO_INT (g_object_get_data (G_OBJECT (source->contents.drive.disc), IS_LOCKED)))
			g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (1));
		else
			g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (0));

		result = bonfire_burn_reload_record_media (burn,
							   BONFIRE_BURN_ERROR_NONE,
							   flags,
							   drive,
							   track,
							   speed,
							   TRUE,
							   error);
		if (result != BONFIRE_BURN_OK)
			goto end;
	}
	else {
		gboolean is_DVD;
		gboolean is_rewritable;

		/* we make sure there is a disc in the drive if it hasn't been done yet */
		result = bonfire_burn_wait_for_recordable_media (burn,
								 flags,
								 drive,
								 source->type,
								 &is_DVD,
								 &is_rewritable,
								 error);
		if (result != BONFIRE_BURN_OK)
			goto end;

		result = bonfire_burn_create_recorder (burn,
						       flags,
						       drive,
						       track,
						       speed,
						       is_rewritable,
						       is_DVD,
						       error);
		if (result != BONFIRE_BURN_OK)
			goto end;
	}

	do {
		int error_code;
		GError *ret_error = NULL;

		result = bonfire_recorder_record (burn->priv->recorder, &ret_error);
		if (result == BONFIRE_BURN_OK)
			break;

		if (result != BONFIRE_BURN_ERR || !ret_error)
			break;

		/* see if error is recoverable */
		error_code = ret_error->code;
		if (error_code == BONFIRE_BURN_ERROR_JOLIET_TREE)
			/* this way error won't be reported if 
			 * user don't want to disable joliet */
			result = bonfire_burn_ask_for_joliet (burn);
		else if (error_code > BONFIRE_BURN_ERROR_DISC_SPACE) {
			result = bonfire_burn_reload_record_media (burn,
								   error_code,
								   flags,
								   drive,
								   track,
								   speed,
								   FALSE,
								   &ret_error);
		}
		else {
			g_propagate_error (error, ret_error);
			break;
		}

		g_error_free (ret_error);
		ret_error = NULL;
	} while (result == BONFIRE_BURN_OK);


end:
	if (track && track != source)
		bonfire_track_source_free (track);

	if (burn->priv->recorder) {
		g_object_unref (burn->priv->recorder);
		burn->priv->recorder = NULL;
	}

	if (burn->priv->imager) {
		g_object_unref (burn->priv->imager);
		burn->priv->imager = NULL;
	}

	if (source->type == BONFIRE_TRACK_SOURCE_DISC && !same_drive) {
		nautilus_burn_drive_unlock (source->contents.drive.disc);
		g_object_set_data (G_OBJECT (source->contents.drive.disc),
				   IS_LOCKED,
				   GINT_TO_POINTER (0));
		if ((flags & BONFIRE_BURN_FLAG_EJECT))
			bonfire_burn_common_eject_async (source->contents.drive.disc);
	}

	if (result == BONFIRE_BURN_OK)
		bonfire_burn_action_changed_real (burn,
						  BONFIRE_BURN_ACTION_FINISHED,
						  NULL);
	else if (result == BONFIRE_BURN_NOT_READY
	     ||  result == BONFIRE_BURN_NOT_SUPPORTED
	     ||  result == BONFIRE_BURN_RUNNING
	     ||  result == BONFIRE_BURN_NOT_RUNNING)
		g_set_error (error,
			     BONFIRE_BURN_ERROR,
			     BONFIRE_BURN_ERROR_GENERAL,
			     _("internal error (code %i)"),
			     result);

	/* we release the drive */
	nautilus_burn_drive_unlock (drive);
	g_object_set_data (G_OBJECT (drive), IS_LOCKED, GINT_TO_POINTER (0));
	if ((flags & BONFIRE_BURN_FLAG_EJECT))
		bonfire_burn_common_eject_async (drive);

	nautilus_burn_drive_unref (burn->priv->drive);
	burn->priv->drive = NULL;

	return result;
}

BonfireBurnResult
bonfire_burn_cancel (BonfireBurn *burn, gboolean protect)
{
	BonfireBurnResult result = BONFIRE_BURN_OK;

	g_return_val_if_fail (burn != NULL, BONFIRE_BURN_ERR);

	if (burn->priv->recorder
	&&  bonfire_job_is_running (BONFIRE_JOB (burn->priv->recorder)))
		result = bonfire_job_cancel (BONFIRE_JOB (burn->priv->recorder), protect);
	else if (burn->priv->imager
	      &&  BONFIRE_JOB (burn->priv->imager) != BONFIRE_JOB (burn->priv->recorder))
		result = bonfire_job_cancel (BONFIRE_JOB (burn->priv->imager), protect);
	else if (burn->priv->recorder)
		result = bonfire_job_cancel (BONFIRE_JOB (burn->priv->recorder), protect);
	else if (burn->priv->imager)
		result = bonfire_job_cancel (BONFIRE_JOB (burn->priv->imager), protect);

	return result;
}
			     
void
bonfire_burn_plug_progress_widget (BonfireBurn *burn,
				   BonfireBurnProgress *progress)
{
	g_return_if_fail (BONFIRE_IS_BURN_PROGRESS (progress));

	if (burn->priv->progress)
		g_object_unref (burn->priv->progress);

	g_object_ref (progress);
	burn->priv->progress = progress;
}
