/***************************************************************************
 *            disc-monitor.c
 *
 *  lun jui 11 11:31:03 2005
 *  Copyright  2005  Rouquier Philippe
 *  bonfire-app@wanadoo.fr
 ***************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <string.h>

#include <glib.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>

#include <libhal.h>

#include <nautilus-burn-drive.h>

#include "disc-monitor.h"

static void bonfire_disc_monitor_class_init (BonfireDiscMonitorClass *klass);
static void bonfire_disc_monitor_init (BonfireDiscMonitor * sp);
static void bonfire_disc_monitor_finalize (GObject * object);

static gboolean bonfire_disc_monitor_hal_init (BonfireDiscMonitor * obj);

static void bonfire_disc_monitor_device_added_cb (LibHalContext * ctx,
						  const char *udi);
static void bonfire_disc_monitor_device_removed_cb (LibHalContext * ctx,
						    const char *udi);

static BonfireDisc *bonfire_disc_monitor_new_disc (BonfireDiscMonitor *monitor,
						   const char *udi);

struct BonfireDiscMonitorPrivate {
	LibHalContext *ctx;
	GSList *discs;
};

typedef enum {
	/* Place Signal Types Here */
	DISC_ADDED,
	DISC_REMOVED,
	LAST_SIGNAL
} BonfireDiscMonitorSignalType;

static guint bonfire_disc_monitor_signals[LAST_SIGNAL] = { 0 };
static GObjectClass *parent_class = NULL;
static BonfireDiscMonitor *main_monitor = NULL;

GType
bonfire_disc_monitor_get_type ()
{
	static GType type = 0;

	if (type == 0) {
		static const GTypeInfo our_info = {
			sizeof (BonfireDiscMonitorClass),
			NULL,
			NULL,
			(GClassInitFunc) bonfire_disc_monitor_class_init,
			NULL,
			NULL,
			sizeof (BonfireDiscMonitor),
			0,
			(GInstanceInitFunc) bonfire_disc_monitor_init,
		};

		type = g_type_register_static (G_TYPE_OBJECT,
					       "BonfireDiscMonitor",
					       &our_info, 0);
	}

	return type;
}

static void
bonfire_disc_monitor_class_init (BonfireDiscMonitorClass * klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);
	object_class->finalize = bonfire_disc_monitor_finalize;

	bonfire_disc_monitor_signals[DISC_ADDED] =
	    g_signal_new ("disc_added", G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_ACTION | G_SIGNAL_RUN_FIRST,
			  G_STRUCT_OFFSET (BonfireDiscMonitorClass,
					   disc_added), NULL, NULL,
			  g_cclosure_marshal_VOID__POINTER, G_TYPE_NONE, 1,
			  G_TYPE_POINTER);
	bonfire_disc_monitor_signals[DISC_REMOVED] =
	    g_signal_new ("disc_removed",
			  G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_ACTION | G_SIGNAL_RUN_FIRST,
			  G_STRUCT_OFFSET (BonfireDiscMonitorClass,
					   disc_removed), NULL, NULL,
			  g_cclosure_marshal_VOID__POINTER, G_TYPE_NONE, 1,
			  G_TYPE_POINTER);
}

static void
bonfire_disc_monitor_init (BonfireDiscMonitor * obj)
{
	obj->priv = g_new0 (BonfireDiscMonitorPrivate, 1);
	bonfire_disc_monitor_hal_init (obj);
}

static gboolean
bonfire_disc_monitor_hal_init (BonfireDiscMonitor * monitor)
{
	DBusError error;
	DBusConnection *dbus_connection;
	int nb_devices, nb_volumes, i, j;
	char **devices = NULL, **volumes = NULL;

	/* initialize the connection with hal */
	monitor->priv->ctx = libhal_ctx_new ();
	if (monitor->priv->ctx == NULL) {
		g_warning ("Cannot initialize hal library\n");
		goto error;
	}

	dbus_error_init (&error);
	dbus_connection = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
	if (dbus_error_is_set (&error)) {
		g_warning ("Cannot connect to DBus %s\n", error.message);
		dbus_error_free (&error);
		goto error;
	}

	dbus_connection_setup_with_g_main (dbus_connection, NULL);
	libhal_ctx_set_dbus_connection (monitor->priv->ctx,
					dbus_connection);

	libhal_ctx_set_device_added (monitor->priv->ctx,
				     bonfire_disc_monitor_device_added_cb);
	libhal_ctx_set_device_removed (monitor->priv->ctx,
				       bonfire_disc_monitor_device_removed_cb);
	libhal_ctx_set_user_data (monitor->priv->ctx, monitor);
	libhal_ctx_set_cache (monitor->priv->ctx, FALSE);

	if (libhal_ctx_init (monitor->priv->ctx, &error) == FALSE) {
		g_warning ("Failed to initialize hal : %s\n",
			   error.message);
		dbus_error_free (&error);
		goto error;
	}

	/* Now we get the list and cache it */
	devices = libhal_find_device_by_capability (monitor->priv->ctx,
						    "storage.cdrom", &nb_devices,
						    &error);
	if (dbus_error_is_set (&error)) {
		g_warning ("Hal is not running : %s\n", error.message);
		dbus_error_free (&error);
		goto error;
	}

	for (i = 0; i < nb_devices; i++) {
		/* check it can write */
		if (!libhal_device_get_property_int (monitor->priv->ctx,
						     devices[i],
						     "storage.cdrom.write_speed",
						     &error)) {

			if (dbus_error_is_set (&error)) {
				g_warning ("Hal connection problem :  %s\n",
					   error.message);
				dbus_error_free (&error);
				goto error;
			}
			continue;
		}

		volumes = libhal_manager_find_device_string_match (monitor->priv->ctx,
								   "info.parent",
								   devices[i],
								   &nb_volumes,
								   &error);
		if (dbus_error_is_set (&error)) {
			g_warning ("Hal connection problem :  %s\n",
				   error.message);
			dbus_error_free (&error);
			goto error;
		}

		for (j = 0; j < nb_volumes; j++)
			bonfire_disc_monitor_new_disc (monitor,
						       volumes[j]);

		libhal_free_string_array (volumes);
	}

	libhal_free_string_array (devices);
	return TRUE;

      error:
	libhal_ctx_shutdown (monitor->priv->ctx, NULL);
	libhal_ctx_free (monitor->priv->ctx);
	monitor->priv->ctx = NULL;

	if (devices)
		libhal_free_string_array (devices);
	if (volumes)
		libhal_free_string_array (volumes);

	return FALSE;
}

static void
bonfire_disc_monitor_device_added_cb (LibHalContext * ctx, const char *udi)
{
	BonfireDiscMonitor *monitor;
	BonfireDisc *disc;

	monitor = libhal_ctx_get_user_data (ctx);

	/* check if it's a blank cd */
	if (libhal_device_property_exists
	    (ctx, udi, "volume.is_disc", NULL) == TRUE
	    && libhal_device_get_property_bool (ctx, udi, "volume.is_disc",
						NULL) == TRUE) {
		disc = bonfire_disc_monitor_new_disc (monitor, udi);
		g_signal_emit (monitor,
			       bonfire_disc_monitor_signals[DISC_ADDED], 0,
			       disc);
	}
}

static void
bonfire_disc_monitor_device_removed_cb (LibHalContext * ctx,
					const char *udi)
{
	BonfireDiscMonitor *monitor;
	BonfireDisc *disc = NULL;
	GSList *iter;

	monitor = libhal_ctx_get_user_data (ctx);

	for (iter = monitor->priv->discs; iter; iter = iter->next) {
		disc = iter->data;
		if (!strcmp (disc->udi, udi)) {
			monitor->priv->discs =
			    g_slist_remove (monitor->priv->discs, disc);
			g_signal_emit (monitor,
				       bonfire_disc_monitor_signals
				       [DISC_REMOVED], 0, disc);
			bonfire_disc_free (disc);
			break;
		}
	}
}

static void
bonfire_disc_monitor_finalize (GObject * object)
{
	BonfireDiscMonitor *cobj;

	cobj = BONFIRE_DISC_MONITOR (object);

	if (cobj->priv->ctx) {
		libhal_ctx_shutdown (cobj->priv->ctx, NULL);
		libhal_ctx_free (cobj->priv->ctx);
	}

	if (cobj->priv->discs) {
		GSList *iter;
		BonfireDisc *disc;

		for (iter = cobj->priv->discs; iter; iter = iter->next) {
			disc = iter->data;
			bonfire_disc_free (disc);
		}

		g_slist_free (cobj->priv->discs);
		cobj->priv->discs = NULL;
	}

	g_free (cobj->priv);
	cobj->priv = NULL;

	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static BonfireDisc *
bonfire_disc_monitor_new_disc (BonfireDiscMonitor * monitor,
			       const char *udi)
{
	BonfireDisc *disc;
	char *device_path;
	char *recorder;
	char *parent;
	BonfireDiscType type = BONFIRE_DISC_TYPE_NONE;

	/* we need to know if the volume is mounted */
	if (libhal_device_property_exists (monitor->priv->ctx,
					   udi,
					   "volume.disc.is_mounted",
					   NULL)
	&&  libhal_device_get_property_bool (monitor->priv->ctx,
					     udi,
					     "volume.disc.is_mounted",
					     NULL))
		type |= BONFIRE_DISC_TYPE_MOUNTED;

	/* check it is blank or rewritable */
	if (libhal_device_property_exists (monitor->priv->ctx,
					   udi,
					   "volume.disc.is_blank",
					   NULL)
	&&  libhal_device_get_property_bool (monitor->priv->ctx,
					     udi,
					     "volume.disc.is_blank",
					     NULL))
		type |= BONFIRE_DISC_TYPE_BLANK;

	if (libhal_device_property_exists (monitor->priv->ctx,
					   udi,
					   "volume.disc.has_audio",
					   NULL)
	&&  libhal_device_get_property_bool (monitor->priv->ctx,
					     udi,
					     "volume.disc.has_audio",
					     NULL))
		type |= BONFIRE_DISC_TYPE_AUDIO;

	if (libhal_device_property_exists (monitor->priv->ctx,
					   udi,
					   "volume.disc.is_rewritable",
					   NULL)
	&&  libhal_device_get_property_bool (monitor->priv->ctx,
					     udi,
					     "volume.disc.is_rewritable",
					     NULL))
		type |= BONFIRE_DISC_TYPE_REWRITABLE;

	if (libhal_device_property_exists (monitor->priv->ctx,
					   udi,
					   "volume.disc.is_appendable",
					   NULL)
	&&  libhal_device_get_property_bool (monitor->priv->ctx,
					     udi,
					     "volume.disc.is_appendable",
					     NULL))
		type |= BONFIRE_DISC_TYPE_APPENDABLE;

	disc = g_new0 (BonfireDisc, 1);
	disc->udi = g_strdup (udi);

	parent = libhal_device_get_property_string (monitor->priv->ctx,
						    udi,
						    "block.storage_device",
						    NULL);
	recorder = libhal_device_get_property_string (monitor->priv->ctx,
						      parent,
						      "info.product",
						      NULL);
	libhal_free_string (parent);

	disc->recorder = g_strdup (recorder);
	libhal_free_string (recorder);

	/* FIXME: when mounted nautilus_burn doesn't return the right size */
	device_path = libhal_device_get_property_string (monitor->priv->ctx,
							 udi,
							 "block.device",
							 NULL);

	/* FIXME: the following function is deprecated */
	disc->size = nautilus_burn_drive_get_media_size_from_path (device_path);
	if (disc->size == NAUTILUS_BURN_MEDIA_SIZE_BUSY
	||  disc->size == NAUTILUS_BURN_MEDIA_SIZE_NA
	||  disc->size == NAUTILUS_BURN_MEDIA_SIZE_UNKNOWN) {
		disc->size = -1;
	}
	libhal_free_string (device_path);

	disc->type = type;
	monitor->priv->discs = g_slist_append (monitor->priv->discs, disc);

	return disc;
}

void
bonfire_disc_free (BonfireDisc * disc)
{
	if (disc->udi)
		g_free (disc->udi);
	if (disc->recorder)
		g_free (disc->recorder);

	g_free (disc);
}

BonfireDiscMonitor *
bonfire_disc_monitor_get_default (void)
{
	if (main_monitor) {

		if (main_monitor->priv->ctx == NULL)
			bonfire_disc_monitor_hal_init (main_monitor);

		g_object_ref (main_monitor);
		return main_monitor;
	}

	main_monitor =
	    BONFIRE_DISC_MONITOR (g_object_new
				  (BONFIRE_TYPE_DISC_MONITOR, NULL));

	return main_monitor;
}

GSList *
bonfire_disc_monitor_lookup_disc (BonfireDiscMonitor * monitor,
				  BonfireDiscType type)
{
	BonfireDisc *disc, *newdisc;
	GSList *discs = NULL, *iter;

	for (iter = monitor->priv->discs; iter; iter = iter->next) {
		disc = iter->data;

		if ((disc->type & type)) {
			newdisc = g_memdup (disc, sizeof (BonfireDisc));
			newdisc->udi = g_strdup (disc->udi);
			newdisc->recorder = g_strdup (disc->recorder);
			discs = g_slist_append (discs, newdisc);
		}
	}

	return discs;
}

GSList *
bonfire_disc_monitor_lookup_disc_by_type (BonfireDiscMonitor * monitor,
					  BonfireDiscType type)
{
	BonfireDisc *disc, *newdisc;
	GSList *discs = NULL, *iter;
	BonfireDiscType newtype;

	for (iter = monitor->priv->discs; iter; iter = iter->next) {
		disc = iter->data;

		newtype = disc->type & type;
		if ((newtype == type)) {
			newdisc = g_memdup (disc, sizeof (BonfireDisc));
			newdisc->udi = g_strdup (disc->udi);
			newdisc->recorder = g_strdup (disc->recorder);
			discs = g_slist_append (discs, newdisc);
		}
	}

	return discs;
}

GSList *
bonfire_disc_monitor_lookup_disc_by_type_strict (BonfireDiscMonitor *
						 monitor,
						 BonfireDiscType type)
{
	BonfireDisc *disc, *newdisc;
	GSList *discs = NULL, *iter;

	for (iter = monitor->priv->discs; iter; iter = iter->next) {
		disc = iter->data;

		if (disc->type == type) {
			newdisc = g_memdup (disc, sizeof (BonfireDisc));
			newdisc->udi = g_strdup (disc->udi);
			newdisc->recorder = g_strdup (disc->recorder);
			discs = g_slist_append (discs, newdisc);
		}
	}

	return discs;
}
