/***************************************************************************
 *            project-size.c
 *
 *  dim nov 27 14:33:46 2005
 *  Copyright  2005  Rouquier Philippe
 *  bonfire-app@wanadoo.fr
 ***************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include <string.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include <glib/gi18n-lib.h>
#include <glib/gstdio.h>

#include <gtk/gtkhbox.h>
#include <gtk/gtkprogressbar.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtkcellrenderer.h>
#include <gtk/gtkcellrenderertext.h>
#include <gtk/gtkcontainer.h>
#include <gtk/gtkcombobox.h>
#include <gtk/gtkalignment.h>
#include <gtk/gtkcelllayout.h>
#include <gtk/gtkdialog.h>
#include <gtk/gtkmessagedialog.h>
#include <gtk/gtkstock.h>

#include "project-size.h"
#include "disc-monitor.h"
#include "utils.h"

static void bonfire_project_size_class_init(BonfireProjectSizeClass *klass);
static void bonfire_project_size_init(BonfireProjectSize *sp);
static void bonfire_project_size_finalize(GObject *object);
static void bonfire_project_size_base_class_finalize (BonfireProjectSizeClass *klass);

static void
bonfire_project_size_add_disc (const char *udi,
			       guint64 size,
			       const char *recorder);

static void
bonfire_project_size_combo_inserted_cb (GtkTreeModel *model,
				        GtkTreePath *path,
				        GtkTreeIter *iter,
				        BonfireProjectSize *size);
static void
bonfire_project_size_combo_removed_cb (GtkTreeModel *model,
				       GtkTreePath *path,
				       BonfireProjectSize *size);
static void
bonfire_project_size_combo_changed_cb (GtkComboBox *combo,
				       BonfireProjectSize *size);

static void
bonfire_project_size_disc_added_cb (BonfireDiscMonitor *monitor,
				    const BonfireDisc *disc,
				    gpointer data);
static void
bonfire_project_size_disc_removed_cb (BonfireDiscMonitor *monitor,
				      const BonfireDisc *disc,
				      gpointer data);

struct BonfireProjectSizePrivate {
	GtkWidget *progress;
	GtkWidget *combo;

	double disc_size;
	double max_size;
	double size;

	char *size_string;
	int refresh_id;
};

static BonfireDiscMonitor *monitor = NULL;
static GtkListStore *devices_store = NULL;

enum {
	BONFIRE_PROJECT_SIZE_COMBO_DISPLAY_COL,
	BONFIRE_PROJECT_SIZE_COMBO_NUM_COL,
	BONFIRE_PROJECT_SIZE_COMBO_UDI_COL,
	BONFIRE_PROJECT_SIZE_COMBO_NB_COL
};

typedef enum {
	DISC_CHANGED_SIGNAL,
	LAST_SIGNAL
} BonfireProjectSizeSignalType;


static guint bonfire_project_size_signals[LAST_SIGNAL] = { 0 };
static GObjectClass *parent_class = NULL;

GType
bonfire_project_size_get_type()
{
	static GType type = 0;

	if(type == 0) {
		static const GTypeInfo our_info = {
			sizeof (BonfireProjectSizeClass),
			NULL,
			(GBaseFinalizeFunc)bonfire_project_size_base_class_finalize,
			(GClassInitFunc)bonfire_project_size_class_init,
			NULL,
			NULL,
			sizeof (BonfireProjectSize),
			0,
			(GInstanceInitFunc)bonfire_project_size_init,
		};

		type = g_type_register_static(GTK_TYPE_HBOX, 
			"BonfireProjectSize", &our_info, 0);
	}

	return type;
}

static void
bonfire_project_size_class_init (BonfireProjectSizeClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);
	GSList *list, *iter;
	BonfireDisc *disc;
	int i;
	double len_array[] = { 650 * 1024 * 1024,
		700 * 1024 * 1024,
		800 * 1024 * 1024,
		900 * 1024 * 1024,
		4.3 * 1024 * 1024 * 1024,
		4.7 * 1024 * 1024 * 1024
	};
	
	parent_class = g_type_class_peek_parent (klass);
	object_class->finalize = bonfire_project_size_finalize;

	/* Initialize the combo model */
	devices_store = gtk_list_store_new (BONFIRE_PROJECT_SIZE_COMBO_NB_COL,
					    G_TYPE_STRING,
					    G_TYPE_DOUBLE,
					    G_TYPE_STRING);

	for (i = 0; i < sizeof (len_array) / sizeof (len_array [0]); i++)
		bonfire_project_size_add_disc (NULL, len_array [i], NULL);

	monitor = bonfire_disc_monitor_get_default ();
	list = bonfire_disc_monitor_lookup_disc (monitor,
						 BONFIRE_DISC_TYPE_BLANK |
					         BONFIRE_DISC_TYPE_REWRITABLE);
	if (list) {
		for (iter = list; iter; iter = iter->next) {
			disc = iter->data;

			bonfire_project_size_add_disc (disc->udi,
						       disc->size,
						       disc->recorder);
			bonfire_disc_free (disc);
		}

		g_slist_free (list);
	}

	/* add a notify callback if states changes */
	g_signal_connect (G_OBJECT (monitor), "disc-added",
			  G_CALLBACK (bonfire_project_size_disc_added_cb),
			  NULL);
	g_signal_connect (G_OBJECT (monitor), "disc-removed",
			  G_CALLBACK (bonfire_project_size_disc_removed_cb),
			  NULL);

	bonfire_project_size_signals[DISC_CHANGED_SIGNAL] =
	    g_signal_new ("disc_changed",
			  G_OBJECT_CLASS_TYPE (object_class),
			  G_SIGNAL_ACTION | G_SIGNAL_RUN_FIRST,
			  G_STRUCT_OFFSET (BonfireProjectSizeClass, disc_changed),
			  NULL, NULL,
			  g_cclosure_marshal_VOID__VOID,
			  G_TYPE_NONE,
			  0);

}

static void
bonfire_project_size_base_class_finalize (BonfireProjectSizeClass *klass)
{
	g_object_unref (devices_store);
	devices_store = NULL;

	g_object_unref (monitor);
	monitor = NULL;
}

static void
bonfire_project_size_init (BonfireProjectSize *obj)
{
	GtkCellRenderer *renderer;

	obj->priv = g_new0(BonfireProjectSizePrivate, 1);
	gtk_box_set_spacing (GTK_BOX (obj), 12);

	/* progress bar */
	obj->priv->progress = gtk_progress_bar_new ();
	gtk_progress_bar_set_text (GTK_PROGRESS_BAR (obj->priv->progress),
				   _("empty"));
	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR
				       (obj->priv->progress), 0.0);

	/* combo box */
	obj->priv->combo = gtk_combo_box_new ();
	gtk_combo_box_set_model (GTK_COMBO_BOX (obj->priv->combo),
				 GTK_TREE_MODEL (devices_store));
	g_signal_connect (G_OBJECT (devices_store),
			  "row-inserted",
			  G_CALLBACK (bonfire_project_size_combo_inserted_cb), obj);
	g_signal_connect (G_OBJECT (devices_store), "row-deleted",
			  G_CALLBACK (bonfire_project_size_combo_removed_cb),
			  obj);
	g_signal_connect (G_OBJECT (obj->priv->combo), "changed",
			  G_CALLBACK (bonfire_project_size_combo_changed_cb),
			  obj);

	renderer = gtk_cell_renderer_text_new ();
	gtk_cell_layout_pack_start (GTK_CELL_LAYOUT (obj->priv->combo),
				    renderer, TRUE);
	gtk_cell_layout_add_attribute (GTK_CELL_LAYOUT (obj->priv->combo),
				       renderer, "text",
				       BONFIRE_PROJECT_SIZE_COMBO_DISPLAY_COL);

	gtk_combo_box_set_active (GTK_COMBO_BOX (obj->priv->combo), 0);

	gtk_box_pack_end (GTK_BOX (obj), obj->priv->combo, FALSE, TRUE, 0);
	gtk_box_pack_end (GTK_BOX (obj), obj->priv->progress, TRUE, TRUE, 0);
}

static void
bonfire_project_size_finalize (GObject *object)
{
	BonfireProjectSize *cobj;
	cobj = BONFIRE_PROJECT_SIZE(object);
	
	if (cobj->priv->refresh_id)
		g_source_remove (cobj->priv->refresh_id);

	if (cobj->priv->size_string)
		g_free (cobj->priv->size_string);

	g_free(cobj->priv);
	G_OBJECT_CLASS(parent_class)->finalize(object);
}

GtkWidget *
bonfire_project_size_new ()
{
	BonfireProjectSize *obj;
	
	obj = BONFIRE_PROJECT_SIZE(g_object_new(BONFIRE_TYPE_PROJECT_SIZE, NULL));
	
	return GTK_WIDGET (obj);
}

/************************************* COMBO STUFF *****************************/
static void
bonfire_project_size_add_disc (const char *udi,
			       guint64 size,
			       const char *recorder)
{
	GtkTreeIter row;
	char *size_string, *time_string, *text;

	/* we should round the disc sizes / length */
	if (size == -1) {
		if (recorder)
			text = g_strdup_printf (_("(mounted drive) %s "), recorder);
		else
			return;
	}
	else {
		time_string = bonfire_utils_get_time_string_from_size (size, TRUE, TRUE);
		size_string = bonfire_utils_get_size_string (size, TRUE, TRUE);

		if (recorder)
			text = g_strdup_printf ("%s / %s (%s)",
						time_string,
						size_string,
						recorder);
		else
			text = g_strdup_printf ("%s / %s", time_string, size_string);

		g_free (size_string);
		g_free (time_string);
	}

	gtk_list_store_prepend (devices_store, &row);
	gtk_list_store_set (devices_store,
			    &row,
			    BONFIRE_PROJECT_SIZE_COMBO_DISPLAY_COL, text,
			    BONFIRE_PROJECT_SIZE_COMBO_NUM_COL,
			    (double) size,
			    BONFIRE_PROJECT_SIZE_COMBO_UDI_COL, udi, -1);
	g_free (text);
}

static void
bonfire_project_size_disc_added_cb (BonfireDiscMonitor *monitor,
				    const BonfireDisc *disc,
				    gpointer data)
{
	if (disc->type & (BONFIRE_DISC_TYPE_BLANK | BONFIRE_DISC_TYPE_REWRITABLE))
		bonfire_project_size_add_disc (disc->udi,
					       disc->size,
					       disc->recorder);
}

static void
bonfire_project_size_disc_removed_cb (BonfireDiscMonitor *monitor,
				      const BonfireDisc *disc,
				      gpointer data)
{
	GtkTreeIter row;
	char *row_udi;

	if (gtk_tree_model_get_iter_first
	    (GTK_TREE_MODEL (devices_store), &row) == FALSE)
		return;

	do {
		gtk_tree_model_get (GTK_TREE_MODEL (devices_store),
				    &row,
				    BONFIRE_PROJECT_SIZE_COMBO_UDI_COL,
				    &row_udi, -1);

		if (!row_udi)
			continue;

		if (!strcmp (row_udi, disc->udi)) {
			gtk_list_store_remove (devices_store, &row);
			g_free (row_udi);
			break;
		}

		g_free (row_udi);
	} while (gtk_tree_model_iter_next
		 (GTK_TREE_MODEL (devices_store), &row) == TRUE);
}
static gboolean
bonfire_project_size_update_size (BonfireProjectSize *obj)
{
	double fraction = 0.0;
	char *string;

	if (!obj->priv->disc_size) {
		fraction = 0.0;
		string = g_strdup_printf (_("no disc chosen (%s)"),
					  obj->priv->size_string);
	}
	else if (obj->priv->size > obj->priv->disc_size) {
		fraction = 1.0;
		string = g_strdup_printf (_("oversized (%s)"),
					  obj->priv->size_string);
	}
	else if (obj->priv->size <= 0.0) {
		fraction = 0.0;
		string = g_strdup (_("empty"));
	}
	else if (obj->priv->size_string
	     &&  obj->priv->size_string [0] != '\0') {
		string = g_strdup (obj->priv->size_string);
		fraction = (double) (obj->priv->size / obj->priv->disc_size);
	}
	else
		string = g_strdup (_("empty"));

	gtk_progress_bar_set_text (GTK_PROGRESS_BAR (obj->priv->progress),
				   string);
	g_free (string);

	gtk_progress_bar_set_fraction (GTK_PROGRESS_BAR (obj->priv->progress),
				       fraction);

	obj->priv->refresh_id = 0;
	return FALSE;
}

static void
bonfire_project_size_refresh (BonfireProjectSize *obj)
{
	if (!obj->priv->refresh_id)
		obj->priv->refresh_id = g_timeout_add (500,
						       (GSourceFunc) bonfire_project_size_update_size,
						       obj);
}

static void
bonfire_project_size_combo_changed (BonfireProjectSize *size)
{
	GtkTreeIter iter;
	double cd_size;

	if (gtk_combo_box_get_active_iter
	    (GTK_COMBO_BOX (size->priv->combo), &iter) == FALSE) {
		/* no default active iter => set one */
		gtk_tree_model_get_iter_first (GTK_TREE_MODEL (devices_store), &iter);
		g_signal_handlers_block_by_func (G_OBJECT (size->priv->combo),
						 bonfire_project_size_combo_changed_cb,
						 size);
		gtk_combo_box_set_active_iter (GTK_COMBO_BOX (size->priv->combo),
					       &iter);
		g_signal_handlers_unblock_by_func (G_OBJECT (size->priv->combo),
						   bonfire_project_size_combo_changed_cb,
						   size);
	}

	gtk_tree_model_get (GTK_TREE_MODEL (devices_store),
			    &iter,
			    BONFIRE_PROJECT_SIZE_COMBO_NUM_COL,
			    &cd_size, -1);


	size->priv->disc_size = cd_size;
	size->priv->max_size = cd_size * 105 / 100;
	bonfire_project_size_refresh (size);

	g_signal_emit (size,
		       bonfire_project_size_signals [DISC_CHANGED_SIGNAL],
		       0);
}

static void
bonfire_project_size_combo_inserted_cb (GtkTreeModel *model,
				        GtkTreePath *path,
				        GtkTreeIter *iter,
				        BonfireProjectSize *size)
{
	bonfire_project_size_combo_changed (size);
}

static void
bonfire_project_size_combo_removed_cb (GtkTreeModel *model,
				       GtkTreePath *path,
				       BonfireProjectSize *size)
{
	bonfire_project_size_combo_changed (size);
}

static void
bonfire_project_size_combo_changed_cb (GtkComboBox *combo,
				       BonfireProjectSize *size)
{
	bonfire_project_size_combo_changed (size);
}

void
bonfire_project_size_set_size (BonfireProjectSize *obj,
			       double size,
			       const char *string)
{
	obj->priv->size = size;
	if (obj->priv->size_string)
		g_free (obj->priv->size_string);

	obj->priv->size_string = g_strdup (string);

	/* we don't update the size right now but in half a second.
	 * when exploring directories size can changed repeatedly
	 * and we don't want to lose too much time updating.
	 * if a size is already set, we know that we're waiting for
	 * a size update, so, just replace the old size. otherwise
	 * we add a g_timeout_add */
	bonfire_project_size_refresh (obj);
}

gboolean
bonfire_project_size_check_status (BonfireProjectSize *size,
				   gboolean *overburn)
{
	if (size->priv->max_size < size->priv->size) {
		*overburn = FALSE;
		return FALSE;
	}

	if (size->priv->disc_size < size->priv->size) {
		*overburn = TRUE;
		return FALSE;
	}

	return TRUE;
}
