/***************************************************************************
 *            project.c
 *
 *  mar nov 29 09:32:17 2005
 *  Copyright  2005  Rouquier Philippe
 *  bonfire-app@wanadoo.fr
 ***************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */



#include <string.h>

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <glib.h>
#include <glib/gstdio.h>
#include <glib/gi18n.h>

#include <gtk/gtknotebook.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtkfilechooser.h>
#include <gtk/gtkfilechooserdialog.h>
#include <gtk/gtkmessagedialog.h>
#include <gtk/gtkprogressbar.h>
#include <gtk/gtkstock.h>

#include <libxml/xmlerror.h>
#include <libxml/xmlwriter.h>
#include <libxml/parser.h>
#include <libxml/xmlstring.h>

#include <libgnomevfs/gnome-vfs.h>
#include <libgnomevfs/gnome-vfs-file-info.h>

#ifdef BUILD_GDL
#include <gdl/gdl-dock.h>
#include <gdl/gdl-dock-item.h>
#endif

#include "project.h"
#include "project-size.h"
#include "project-type-chooser.h"
#include "disc.h"
#include "data-disc.h"
#include "audio-disc.h"
#include "burn-options-dialog.h"
#include "disc-copy-dialog.h"
#include "utils.h"


static void bonfire_project_class_init(BonfireProjectClass *klass);
static void bonfire_project_init(BonfireProject *sp);
static void bonfire_project_finalize(GObject *object);

static void
bonfire_project_disc_changed_cb (BonfireProjectSize *size,
				 BonfireProject *project);
static void
bonfire_project_size_changed_cb (BonfireDisc *disc,
			         double size,
			         const char *string,
			         BonfireProject *project);
static void
bonfire_project_selection_changed (BonfireDisc *disc,
				   gboolean empty,
				   BonfireProject *project);
static void
bonfire_project_chosen_cb (BonfireProjectTypeChooser *chooser,
			   BonfireProjectType type,
			   BonfireProject *project);

struct BonfireProjectPrivate {
	GtkWidget *size_display;
	GtkWidget *discs;
	GtkWidget *audio;
	GtkWidget *data;
	GtkWidget *chooser;

	char *project;
	BonfireProjectType type;

	int oversized:1;
	int overburn:1;
	int ask_overburn:1;
};

enum {
	SELECTION_CHANGED_SIGNAL,
	LAST_SIGNAL
};

guint bonfire_project_signals[LAST_SIGNAL];
static GObjectClass *parent_class = NULL;

#define BONFIRE_PROJECT_VERSION "0.2"

GType
bonfire_project_get_type ()
{
	static GType type = 0;

	if(type == 0) {
		static const GTypeInfo our_info = {
			sizeof (BonfireProjectClass),
			NULL,
			NULL,
			(GClassInitFunc)bonfire_project_class_init,
			NULL,
			NULL,
			sizeof (BonfireProject),
			0,
			(GInstanceInitFunc)bonfire_project_init,
		};

		type = g_type_register_static(GTK_TYPE_NOTEBOOK, 
			"BonfireProject", &our_info, 0);
	}

	return type;
}

static void
bonfire_project_class_init (BonfireProjectClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);

	parent_class = g_type_class_peek_parent(klass);
	object_class->finalize = bonfire_project_finalize;
	
	bonfire_project_signals[SELECTION_CHANGED_SIGNAL] =
	    g_signal_new ("selection_changed",
			  G_TYPE_FROM_CLASS (klass),
			  G_SIGNAL_RUN_LAST,
			  G_STRUCT_OFFSET (BonfireProjectClass, selection_changed),
			  NULL, NULL,
			  g_cclosure_marshal_VOID__BOOLEAN,
			  G_TYPE_NONE,
			  1,
			  G_TYPE_BOOLEAN);
}

static void
bonfire_project_init (BonfireProject *obj)
{
	GtkWidget *scroll;
	GtkWidget *box;

	obj->priv = g_new0 (BonfireProjectPrivate, 1);
	gtk_container_set_border_width (GTK_CONTAINER (obj), 8);

	obj->priv->size_display = bonfire_project_size_new ();
	g_signal_connect (G_OBJECT (obj->priv->size_display), 
			  "disc-changed",
			  G_CALLBACK (bonfire_project_disc_changed_cb),
			  obj);

	obj->priv->audio = bonfire_audio_disc_new ();
	obj->priv->data = bonfire_data_disc_new ();

	g_signal_connect (G_OBJECT (obj->priv->audio),
			  "selection_changed",
			  G_CALLBACK (bonfire_project_selection_changed), obj);
	g_signal_connect (G_OBJECT (obj->priv->audio),
			  "size_changed",
			  G_CALLBACK (bonfire_project_size_changed_cb), obj);

	g_signal_connect (G_OBJECT (obj->priv->data),
			  "selection_changed",
			  G_CALLBACK (bonfire_project_selection_changed), obj);
	g_signal_connect (G_OBJECT (obj->priv->data),
			  "size_changed",
			  G_CALLBACK (bonfire_project_size_changed_cb), obj);

	obj->priv->discs = gtk_notebook_new ();
	gtk_notebook_set_show_border (GTK_NOTEBOOK (obj->priv->discs), FALSE);
	gtk_notebook_set_show_tabs (GTK_NOTEBOOK (obj->priv->discs), FALSE);
	gtk_notebook_prepend_page (GTK_NOTEBOOK (obj->priv->discs),
				   obj->priv->data, NULL);
	gtk_notebook_prepend_page (GTK_NOTEBOOK (obj->priv->discs),
				   obj->priv->audio, NULL);

	box = gtk_vbox_new (FALSE, 6);
	gtk_box_pack_end (GTK_BOX (box), obj->priv->size_display, FALSE, FALSE, 0);
	gtk_box_pack_end (GTK_BOX (box), obj->priv->discs, TRUE, TRUE, 0);

	obj->priv->chooser = bonfire_project_type_chooser_new ();
	scroll = gtk_scrolled_window_new (NULL, NULL);
	gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (scroll),
					GTK_POLICY_AUTOMATIC,
					GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_add_with_viewport (GTK_SCROLLED_WINDOW (scroll),
					       obj->priv->chooser);
	gtk_notebook_prepend_page (GTK_NOTEBOOK (obj), box, NULL);
	gtk_notebook_prepend_page (GTK_NOTEBOOK (obj), scroll, NULL);

	gtk_widget_show_all (scroll);
	gtk_widget_show (obj->priv->audio);
	gtk_widget_show (obj->priv->data);

	g_signal_connect (G_OBJECT (obj->priv->chooser), "chosen",
			  G_CALLBACK (bonfire_project_chosen_cb), obj);

	gtk_notebook_set_show_border (GTK_NOTEBOOK (obj), FALSE);
	gtk_notebook_set_show_tabs (GTK_NOTEBOOK (obj), FALSE);
}

static void
bonfire_project_finalize (GObject *object)
{
	BonfireProject *cobj;
	cobj = BONFIRE_PROJECT(object);

	if (cobj->priv->project)
		g_free (cobj->priv->project);

	g_free(cobj->priv);
	G_OBJECT_CLASS(parent_class)->finalize(object);
}

GtkWidget *
bonfire_project_new ()
{
	BonfireProject *obj;
	
	obj = BONFIRE_PROJECT(g_object_new(BONFIRE_TYPE_PROJECT, NULL));
	gtk_notebook_set_current_page (GTK_NOTEBOOK (obj), 0);
	return GTK_WIDGET (obj);
}

/********************************** size ***************************************/
static void
bonfire_project_error_size_dialog (BonfireProject *project)
{
	GtkWidget *dialog;
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	dialog = gtk_message_dialog_new (GTK_WINDOW (toplevel),
					 GTK_DIALOG_DESTROY_WITH_PARENT |
					 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_ERROR,
					 GTK_BUTTONS_CLOSE,
					 _("The size of the project is too large for the disc even with the overburn option:"));

	gtk_window_set_title (GTK_WINDOW (dialog), _("Project size"));

	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
						  _("you must delete some files."));

	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}

static void
bonfire_project_overburn_dialog (BonfireProject *project)
{
	GtkWidget *dialog, *toplevel;
	gint result;

	/* get the current CD length and make sure selection is not too long */
	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	dialog = gtk_message_dialog_new (GTK_WINDOW (toplevel),
					 GTK_DIALOG_DESTROY_WITH_PARENT|
					 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_WARNING,
					 GTK_BUTTONS_NONE,
					 _("The size of the project is too large for the disc:"));

	gtk_window_set_title (GTK_WINDOW (dialog), _("Project size"));

	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
						  _("Would you like to activate overburn (otherwise you must delete files) ?"
						    "\nNOTE: This option might cause failure."));

	gtk_dialog_add_buttons (GTK_DIALOG (dialog),
				GTK_STOCK_NO,
				GTK_RESPONSE_NO,
				GTK_STOCK_YES,
				GTK_RESPONSE_YES, NULL);

	result = gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
	if (result == GTK_RESPONSE_YES) {
		project->priv->oversized = 0;
		project->priv->overburn = 1;
	}
	else {
		project->priv->oversized = 1;
		project->priv->overburn = 0;
	}
}

static void
bonfire_project_check_size (BonfireProject *project)
{
	gboolean result;
	gboolean overburn;

	result = bonfire_project_size_check_status (BONFIRE_PROJECT_SIZE (project->priv->size_display),
						    &overburn);

	if (result) {
		project->priv->ask_overburn = 0;
		project->priv->oversized = 0;
		project->priv->overburn = 0;
		goto end;
	}

	if (overburn
	&& !project->priv->ask_overburn) {
		if (!project->priv->overburn)
			bonfire_project_overburn_dialog (project);

		project->priv->ask_overburn = 1;
		goto end;
	}

	if (!overburn
	&&  project->priv->overburn)
		project->priv->overburn = 0;

	/* avoid telling the user the same thing twice */
	if (project->priv->oversized)
		goto end;

	bonfire_project_error_size_dialog (project);
	project->priv->oversized = 1;
	project->priv->overburn = 0;

end:
	if (project->priv->oversized) {
		g_object_set (G_OBJECT (project->priv->data), "reject-file", TRUE, NULL);
		g_object_set (G_OBJECT (project->priv->data), "reject-file", TRUE, NULL);
	}
	else {
		g_object_set (G_OBJECT (project->priv->data), "reject-file", FALSE, NULL);
		g_object_set (G_OBJECT (project->priv->data), "reject-file", FALSE, NULL);
	}
}

static void
bonfire_project_size_changed_cb (BonfireDisc *disc,
			         double size,
			         const char *string,
			         BonfireProject *project)
{
	bonfire_project_size_set_size (BONFIRE_PROJECT_SIZE (project->priv->size_display),
				       size,
				       string);

	bonfire_project_check_size (project);
}

static void
bonfire_project_disc_changed_cb (BonfireProjectSize *size,
				 BonfireProject *project)
{
	bonfire_project_check_size (project);
}

/******************************** useful function to wait **********************/
static gboolean
_wait_for_ready_state (GtkWidget *dialog) {
	GtkProgressBar *progress;
	BonfireProject *project;
	BonfireDisc *disc = NULL;

	project = g_object_get_data (G_OBJECT (dialog), "Project");
	if (project->priv->oversized) {
		gtk_dialog_response (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL);
		return FALSE;
	}

	progress = g_object_get_data (G_OBJECT (dialog), "ProgressBar");

	switch (project->priv->type) {
	case BONFIRE_PROJECT_TYPE_AUDIO:
		disc = BONFIRE_DISC (project->priv->audio);
		break;
	case BONFIRE_PROJECT_TYPE_DATA:
		disc = BONFIRE_DISC (project->priv->data);
		break;
	default :
		gtk_dialog_response (GTK_DIALOG (dialog), GTK_RESPONSE_CANCEL);
		return FALSE;
	}

	if (bonfire_disc_get_status (disc) == BONFIRE_DISC_NOT_READY) {
		gtk_progress_bar_pulse (progress);
		return TRUE;
	}

	gtk_dialog_response (GTK_DIALOG (dialog), GTK_RESPONSE_OK);
	return FALSE;
}

static BonfireDiscResult
bonfire_project_check_status (BonfireProject *project,
			      BonfireDisc *disc)
{
	int id;
	int answer;
	GtkWidget *dialog;
	GtkWidget *progress;
	GtkWidget *toplevel;
	BonfireDiscResult result;

	if (project->priv->oversized) {
		bonfire_project_error_size_dialog (project);
		return BONFIRE_DISC_ERROR_SIZE;
	}

	result = bonfire_disc_get_status (disc);
	if (result != BONFIRE_DISC_NOT_READY)
		return result;

	/* we are not ready to create tracks presumably because
	 * data or audio has not finished to explore a directory
	 * or get the metadata of a song or a film */
	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	dialog = gtk_message_dialog_new (GTK_WINDOW (toplevel),
					 GTK_DIALOG_DESTROY_WITH_PARENT |
					 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_INFO,
					 GTK_BUTTONS_CLOSE,
					 _("Please wait:"));

	gtk_window_set_title (GTK_WINDOW (dialog), _("Please wait"));

	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
						  _("some tasks are not completed yet."));

	gtk_window_set_title (GTK_WINDOW (dialog), _("Waiting for ongoing tasks"));

	progress = gtk_progress_bar_new ();
	gtk_progress_bar_set_text (GTK_PROGRESS_BAR (progress), " ");
	gtk_box_pack_end (GTK_BOX (GTK_DIALOG (dialog)->vbox),
			  progress,
			  TRUE,
			  TRUE,
			  10);

	gtk_widget_show_all (dialog);
	gtk_progress_bar_pulse (GTK_PROGRESS_BAR (progress));

	g_object_set_data (G_OBJECT (dialog), "ProgressBar", progress);
	g_object_set_data (G_OBJECT (dialog), "Project", project);

	id = g_timeout_add (100,
			    (GSourceFunc) _wait_for_ready_state,
		            dialog);

	answer = gtk_dialog_run (GTK_DIALOG (dialog));
	g_source_remove (id);

	gtk_widget_destroy (dialog);

	if (answer != GTK_RESPONSE_OK)
		return BONFIRE_DISC_CANCELLED;
	else if (project->priv->oversized)
		return BONFIRE_DISC_ERROR_SIZE;

	return bonfire_disc_get_status (disc);
}

/********************************     ******************************************/
/*static void
bonfire_project_show_main (GtkObject *object, GtkWidget *window)
{
	gtk_widget_hide (window);
}
*/
static void
bonfire_project_create_burn_dialog (BonfireProject *project,
				    BonfireTrackSource *track)
{
	GtkWidget *toplevel;
	GtkWidget *dialog;

	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	dialog = bonfire_burn_option_dialog_new (track, project->priv->overburn);

	gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (toplevel));
	gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
	gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER_ON_PARENT);

	gtk_widget_show_all (dialog);
}

static void
bonfire_project_burn_iso (BonfireProject *project)
{
	BonfireTrackSource *track;
	GnomeVFSFileInfo *info;
	GnomeVFSResult result;
	char *escaped_uri;
	char *uri;

	if (!project->priv->project) {
		GtkWidget *toplevel;
		GtkWidget *dialog;
		int answer;

		toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
		dialog = gtk_file_chooser_dialog_new (_("Choose a disc image or a cue file"),
						      GTK_WINDOW (toplevel),
						      GTK_FILE_CHOOSER_ACTION_OPEN,
						      GTK_STOCK_CANCEL, GTK_RESPONSE_REJECT,
						      GTK_STOCK_OPEN, GTK_RESPONSE_OK,
						      NULL);

		gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (dialog), FALSE);
		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog),
						     g_get_home_dir ());
	
		gtk_widget_show (dialog);
		answer = gtk_dialog_run (GTK_DIALOG (dialog));

		if (answer != GTK_RESPONSE_OK) {
			gtk_widget_destroy (dialog);
			project->priv->type = BONFIRE_PROJECT_TYPE_INVALID;
			return;
		}

		uri = gtk_file_chooser_get_uri (GTK_FILE_CHOOSER (dialog));
		gtk_widget_destroy (dialog);
	}
	else {
		uri = project->priv->project;
		project->priv->project = NULL;
	}

	/* check if it is an iso and if it is not a remote file */
	info = gnome_vfs_file_info_new ();
	escaped_uri = gnome_vfs_escape_host_and_path_string (uri);
	result = gnome_vfs_get_file_info (escaped_uri,
					  info,
					  GNOME_VFS_FILE_INFO_GET_MIME_TYPE|
					  GNOME_VFS_FILE_INFO_FORCE_SLOW_MIME_TYPE);
	g_free (escaped_uri);

	if (result != GNOME_VFS_OK)
		goto error;

	track = g_new0 (BonfireTrackSource, 1);
	if (!strcmp (info->mime_type, "application/x-cd-image")) {
		track->type = BONFIRE_TRACK_SOURCE_ISO;
		track->contents.iso.image = gnome_vfs_get_local_path_from_uri (uri);
	}
	else if (!strcmp (info->mime_type, "text/plain")) {
		track->type = BONFIRE_TRACK_SOURCE_CUE;
		track->contents.cue.toc = gnome_vfs_get_local_path_from_uri (uri);
	}
	else 
		goto error;

	g_free (uri);
	gnome_vfs_file_info_clear (info);
	gnome_vfs_file_info_unref (info);

	bonfire_project_create_burn_dialog (project, track);
	return;


error:
	{
		GtkWidget *dialog;
		GtkWidget *toplevel;

		gnome_vfs_file_info_clear (info);
		gnome_vfs_file_info_unref (info);

		toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
		dialog = gtk_message_dialog_new (GTK_WINDOW (toplevel),
						 GTK_DIALOG_DESTROY_WITH_PARENT |
						 GTK_DIALOG_MODAL,
						 GTK_MESSAGE_ERROR,
						 GTK_BUTTONS_CLOSE,
						 _("This image can't be burnt:"));

		gtk_window_set_title (GTK_WINDOW (dialog), _("Burning failure"));

		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
							  _("it doesn't appear to be a valid image or cue file."));
	
		gtk_dialog_run (GTK_DIALOG (dialog));
		gtk_widget_destroy (dialog);

	}
}

static void
bonfire_project_switch (BonfireProject *project,
			BonfireProjectType type)
{
	GtkWidget *toplevel;

	if (project->priv->type > BONFIRE_PROJECT_TYPE_ISO) {
		if (project->priv->type == BONFIRE_PROJECT_TYPE_AUDIO)
			bonfire_disc_reset (BONFIRE_DISC (project->priv->audio));
		else if (project->priv->type == BONFIRE_PROJECT_TYPE_DATA)
			bonfire_disc_reset (BONFIRE_DISC (project->priv->data));

		if (project->priv->project) {
			g_free (project->priv->project);
			project->priv->project = NULL;
		}
	}

#ifdef BUILD_GDL
	GdlDock *dock;
	dock = gdl_dock_object_get_toplevel (GDL_DOCK_OBJECT (GTK_WIDGET (project)->parent));
	if (dock)
		toplevel = gtk_widget_get_toplevel (GTK_WIDGET (dock));
	else
		toplevel = NULL;
#else
	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
#endif

	if (type == BONFIRE_PROJECT_TYPE_INVALID) {
		gtk_notebook_set_current_page (GTK_NOTEBOOK (project), 0);
		if (toplevel)
			gtk_window_set_title (GTK_WINDOW (toplevel), "Bonfire");
	}
	else if (type == BONFIRE_PROJECT_TYPE_AUDIO) {
		gtk_notebook_set_current_page (GTK_NOTEBOOK (project), 1);
		gtk_notebook_set_current_page (GTK_NOTEBOOK (project->priv->discs),
					       0);
		if (toplevel)
			gtk_window_set_title (GTK_WINDOW (toplevel),
					      _("Bonfire - New audio disc project"));
	}
	else if (type == BONFIRE_PROJECT_TYPE_DATA) {
		gtk_notebook_set_current_page (GTK_NOTEBOOK (project), 1);
		gtk_notebook_set_current_page (GTK_NOTEBOOK (project->priv->discs),
					       1);
		if (toplevel)
			gtk_window_set_title (GTK_WINDOW (toplevel),
					      _("Bonfire - New data disc project"));
	}
	else if (type == BONFIRE_PROJECT_TYPE_ISO) {
		gtk_notebook_set_current_page (GTK_NOTEBOOK (project), 0);
		if (toplevel)
			gtk_window_set_title (GTK_WINDOW (toplevel),
					      _("Bonfire - New image file"));
		bonfire_project_burn_iso (project);
	}
	else if (type == BONFIRE_PROJECT_TYPE_COPY) {
		GtkWidget *dialog;

		gtk_notebook_set_current_page (GTK_NOTEBOOK (project), 0);
		if (toplevel)
			gtk_window_set_title (GTK_WINDOW (toplevel),
					      _("Bonfire - Copy a disc"));

		dialog = bonfire_disc_copy_dialog_new ();
	
		gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (toplevel));
		gtk_window_set_modal (GTK_WINDOW (dialog), TRUE);
		gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_CENTER_ON_PARENT);

		gtk_widget_show_all (dialog);
	}

	project->priv->type = type;
}

void
bonfire_project_new_project (BonfireProject *project)
{
	bonfire_project_size_set_size (BONFIRE_PROJECT_SIZE (project->priv->size_display),
				       0.0,
				       NULL);

	bonfire_project_switch (project, BONFIRE_PROJECT_TYPE_INVALID);
}

void
bonfire_project_new_audio_project (BonfireProject *project)
{
	bonfire_project_switch (project, BONFIRE_PROJECT_TYPE_AUDIO);
}

void
bonfire_project_new_data_project (BonfireProject *project)
{
	bonfire_project_switch (project, BONFIRE_PROJECT_TYPE_DATA);
}

void
bonfire_project_new_copy_disc (BonfireProject *project)
{
	bonfire_project_switch (project, BONFIRE_PROJECT_TYPE_COPY);
}

void
bonfire_project_new_iso (BonfireProject *project, const char *uri)
{

	if (project->priv->project)
		g_free (project->priv->project);
	project->priv->project = g_strdup (uri);

	bonfire_project_switch (project, BONFIRE_PROJECT_TYPE_ISO);
}

static void
bonfire_project_chosen_cb (BonfireProjectTypeChooser *chooser,
			   BonfireProjectType type,
			   BonfireProject *project)
{
	bonfire_project_switch (project, type);
}

static void
bonfire_project_selection_changed (BonfireDisc *disc,
				   gboolean empty,
				   BonfireProject *project)
{
	g_signal_emit (project,
		       bonfire_project_signals [SELECTION_CHANGED_SIGNAL], 0,
		       empty);
}

char *
bonfire_project_get_content_icon (BonfireProject *project)
{
	switch (project->priv->type) {
	case BONFIRE_PROJECT_TYPE_AUDIO:
		return g_strdup ("gnome-dev-cdrom-audio");
	case BONFIRE_PROJECT_TYPE_DATA:
	default:
		return g_strdup ("gnome-dev-cdrom");
	}
}

void
bonfire_project_add_uris (BonfireProject *project,
			  char **uris)
{
	BonfireDisc *disc;
	char **uri;

	if (!uris)
		return;

	switch (project->priv->type) {
	case BONFIRE_PROJECT_TYPE_AUDIO:
		disc = BONFIRE_DISC (project->priv->audio);
		break;
	case BONFIRE_PROJECT_TYPE_DATA:
		disc = BONFIRE_DISC (project->priv->data);
		break;
	default:
		return;
	}

	uri = uris;
	while (*uri) {
		bonfire_disc_add_uri (disc, *uri);
		uri ++;
	}
}

void
bonfire_project_remove_selected_uris (BonfireProject *project)
{
	switch (project->priv->type) {
	case BONFIRE_PROJECT_TYPE_AUDIO:
		bonfire_disc_delete_selected (BONFIRE_DISC (project->priv->audio));
		break;
	case BONFIRE_PROJECT_TYPE_DATA:
		bonfire_disc_delete_selected (BONFIRE_DISC (project->priv->data));
		break;
	default:
		break;
	}
}

void
bonfire_project_clear (BonfireProject *project)
{
	switch (project->priv->type) {
	case BONFIRE_PROJECT_TYPE_AUDIO:
		bonfire_disc_clear (BONFIRE_DISC (project->priv->audio));
		break;
	case BONFIRE_PROJECT_TYPE_DATA:
		bonfire_disc_clear (BONFIRE_DISC (project->priv->data));
		break;
	default:
		break;
	}
}

static void
bonfire_project_no_song_dialog (BonfireProject *project)
{
	GtkWidget *message;
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	message = gtk_message_dialog_new (GTK_WINDOW (toplevel),
					  GTK_DIALOG_DESTROY_WITH_PARENT|
					  GTK_DIALOG_MODAL,
					  GTK_MESSAGE_WARNING,
					  GTK_BUTTONS_CLOSE,
					  _("No song in the project:"));

	gtk_window_set_title (GTK_WINDOW (message), _("Empty project"));

	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (message),
						  _("please add songs to the project."));

	gtk_dialog_run (GTK_DIALOG (message));
	gtk_widget_destroy (message);
}

static void
bonfire_project_no_file_dialog (BonfireProject *project)
{
	GtkWidget *message;
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	message = gtk_message_dialog_new (GTK_WINDOW (toplevel),
					  GTK_DIALOG_DESTROY_WITH_PARENT|
					  GTK_DIALOG_MODAL,
					  GTK_MESSAGE_WARNING,
					  GTK_BUTTONS_CLOSE,
					  _("Project is empty:"));

	gtk_window_set_title (GTK_WINDOW (message), _("Empty project"));

	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (message),
						  _("please add files to the project."));

	gtk_dialog_run (GTK_DIALOG (message));
	gtk_widget_destroy (message);
}

void
bonfire_project_burn_contents (BonfireProject *project)
{
	BonfireDiscResult result;
	BonfireDiscTrack *track;
	BonfireDisc *current;

	switch (project->priv->type) {
	case BONFIRE_PROJECT_TYPE_AUDIO:
		current = BONFIRE_DISC (project->priv->audio);
		break;
	case BONFIRE_PROJECT_TYPE_DATA:
		current = BONFIRE_DISC (project->priv->data);
		break;
	default:
		return;
	}

	result = bonfire_project_check_status (project, current);
	if (result != BONFIRE_DISC_OK)
		return;

	track = g_new0 (BonfireDiscTrack, 1);
	track->type = BONFIRE_DISC_TRACK_SOURCE;

	result = bonfire_disc_get_track (current, track);
	if (result == BONFIRE_DISC_CANCELLED)
		return;

	if (result == BONFIRE_DISC_ERROR_SIZE)
		return;

	if (!track->contents.src) {
		if (BONFIRE_IS_AUDIO_DISC (current))
			bonfire_project_no_song_dialog (project);
		else
			bonfire_project_no_file_dialog (project);
		return;
	}

	bonfire_project_create_burn_dialog (project, track->contents.src);
	g_free (track);
}

/******************************* Projects **************************************/
static void
bonfire_project_invalid_project_dialog (BonfireProject *project,
					const char *reason)
{
	GtkWidget *dialog;
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	dialog = gtk_message_dialog_new (GTK_WINDOW (toplevel),
					 GTK_DIALOG_DESTROY_WITH_PARENT |
					 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_ERROR,
					 GTK_BUTTONS_CLOSE,
					 _("Error while loading the project:"));

	gtk_window_set_title (GTK_WINDOW (dialog), _("Project loading error"));

	gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
						  reason);

	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}

static BonfireGraftPt *
_read_graft_point (xmlDocPtr project,
		   xmlNodePtr graft)
{
	BonfireGraftPt *retval;

	retval = g_new0 (BonfireGraftPt, 1);
	while (graft) {
		if (!xmlStrcmp (graft->name, (const xmlChar *) "uri")) {
			if (retval->uri)
				goto error;

			retval->uri = (char*) xmlNodeListGetString (project,
							    graft->xmlChildrenNode,
							    1);
			if (!retval->uri)
				goto error;
		}
		else if (!xmlStrcmp (graft->name, (const xmlChar *) "path")) {
			if (retval->path)
				goto error;

			retval->path = (char *) xmlNodeListGetString (project,
								      graft->xmlChildrenNode,
								      1);
			if (!retval->path)
				goto error;
		}
		else if (!xmlStrcmp (graft->name, (const xmlChar *) "excluded")) {
			xmlChar *excluded;

			excluded = xmlNodeListGetString (project,
							 graft->xmlChildrenNode,
							 1);
			if (!excluded)
				goto error;

			retval->excluded = g_slist_prepend (retval->excluded,
							    excluded);
		}
		else if (graft->type == XML_ELEMENT_NODE)
			goto error;

		graft = graft->next;
	}

	retval->excluded = g_slist_reverse (retval->excluded);
	return retval;

error:
	bonfire_graft_point_free (retval);
	return NULL;
}

static BonfireDiscTrack *
_read_data_track (xmlDocPtr project,
		  xmlNodePtr item)
{
	BonfireDiscTrack *track;

	track = g_new0 (BonfireDiscTrack, 1);
	track->type = BONFIRE_DISC_TRACK_DATA;

	while (item) {
		if (!xmlStrcmp (item->name, (const xmlChar *) "graft")) {
			BonfireGraftPt *graft;

			graft = _read_graft_point (project, item->xmlChildrenNode);
			if (!graft)
				goto error;

			track->contents.data.grafts = g_slist_prepend (track->contents.data.grafts, graft);
		}
		else if (!xmlStrcmp (item->name, (const xmlChar *) "restored")) {
			xmlChar *restored;

			restored = xmlNodeListGetString (project,
							 item->xmlChildrenNode,
							 1);
			if (!restored)
				goto error;

			track->contents.data.restored = g_slist_prepend (track->contents.data.restored, restored);
		}
		else if (item->type == XML_ELEMENT_NODE)
			goto error;

		item = item->next;
	}

	track->contents.data.grafts = g_slist_reverse (track->contents.data.grafts);
	return (BonfireDiscTrack *) track;

error :
	bonfire_track_free ((BonfireDiscTrack*) track);
	return NULL;
}

static BonfireDiscTrack *
_read_audio_track (xmlDocPtr project,
		   xmlNodePtr uris)
{
	BonfireDiscTrack *track;

	track = g_new0 (BonfireDiscTrack, 1);
	track->type = BONFIRE_DISC_TRACK_AUDIO;

	while (uris) {
		if (!xmlStrcmp (uris->name, (const xmlChar *) "uri")) {
			char *uri;

			uri = (char *) xmlNodeListGetString (project,
						    uris->xmlChildrenNode,
						    1);
			if (!uri)
				goto error;

			track->contents.uris = g_slist_prepend (track->contents.uris, uri);
		}
		else if (uris->type == XML_ELEMENT_NODE)
			goto error;

		uris = uris->next;
	}

	track->contents.uris = g_slist_reverse (track->contents.uris);
	return (BonfireDiscTrack*) track;

error:
	bonfire_track_free ((BonfireDiscTrack *) track);
	return NULL;
}

static gboolean
_get_tracks (xmlDocPtr project,
	     xmlNodePtr track_node,
	     BonfireDiscTrack **track)
{
	BonfireDiscTrack *newtrack;

	track_node = track_node->xmlChildrenNode;

	newtrack = NULL;
	while (track_node) {
		if (!xmlStrcmp (track_node->name, (const xmlChar *) "audio")) {
			if (newtrack)
				goto error;

			newtrack = _read_audio_track (project,
						      track_node->xmlChildrenNode);
			if (!newtrack)
				goto error;
		}
		else if (!xmlStrcmp (track_node->name, (const xmlChar *) "data")) {
			if (newtrack)
				goto error;

			newtrack = _read_data_track (project,
						     track_node->xmlChildrenNode);

			if (!newtrack)
				goto error;
		}
		else if (track_node->type == XML_ELEMENT_NODE)
			goto error;

		track_node = track_node->next;
	}

	if (!newtrack)
		goto error;

	*track = newtrack;
	return TRUE;

error :
	if (newtrack)
		bonfire_track_free (newtrack);

	bonfire_track_free (newtrack);
	return FALSE;
}

static gboolean
bonfire_project_open_project_xml (BonfireProject *proj,
				  const char *uri,
				  BonfireDiscTrack **track)
{
	xmlNodePtr track_node = NULL;
	xmlDocPtr project;
	xmlChar *version;
	xmlNodePtr item;
	gboolean retval;

	/* start parsing xml doc */
	project = xmlParseFile (uri);
	if (!project) {
		bonfire_project_invalid_project_dialog (proj,
							_("the project could not be opened."));
		return FALSE;
	}

	/* parses the "header" */
	item = xmlDocGetRootElement (project);
	if (!item) {
		bonfire_project_invalid_project_dialog (proj,
							_("the file is empty."));
		return FALSE;
	}

	version = NULL;
	if (xmlStrcmp (item->name, (const xmlChar *) "bonfireproject")
	||  item->next)
		goto error;

	item = item->children;
	while (item) {
		if (!xmlStrcmp (item->name, (const xmlChar *) "version")) {
			if (version)
				goto error;

			version =  xmlNodeListGetString (project,
							 item->xmlChildrenNode,
							 1);
		}
		else if (!xmlStrcmp (item->name, (const xmlChar *) "track")) {
			if (track_node)
				goto error;

			track_node = item;
		}
		else if (item->type == XML_ELEMENT_NODE)
			goto error;


		item = item->next;
	}

	if (!version)
		goto error;

	if (strcmp ((char*) version, BONFIRE_PROJECT_VERSION))
		goto error;

	xmlFree (version);
	retval = _get_tracks (project, track_node, track);
	xmlFreeDoc (project);

	if (!retval)
		bonfire_project_invalid_project_dialog (proj,
							_("it doesn't seem to be a valid bonfire project."));

	return retval;

error:

	if (version)
		xmlFree (version);

	xmlFreeDoc (project);
	bonfire_project_invalid_project_dialog (proj,
						_("it doesn't seem to be a valid bonfire project."));
	return FALSE;
}

BonfireDiscResult
bonfire_project_open_project (BonfireProject *project,
			      const char *name)
{
	BonfireDiscTrack *track = NULL;
	BonfireDiscResult result;
	GtkWidget *toplevel;
	BonfireDisc *disc;
	char *path;

#ifdef BUILD_GDL
	GdlDock *dock;
	dock = gdl_dock_object_get_toplevel (GDL_DOCK_OBJECT (GTK_WIDGET (project)->parent));
	if (dock)
		toplevel = gtk_widget_get_toplevel (GTK_WIDGET (dock));
	else
		toplevel = NULL;
#else
	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
#endif

	if (!name) {
		GtkWidget *chooser;
		int answer;

		chooser = gtk_file_chooser_dialog_new (_("Open a project"),
						      GTK_WINDOW (toplevel),
						      GTK_FILE_CHOOSER_ACTION_OPEN,
						      GTK_STOCK_OK, GTK_RESPONSE_OK,
						      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
						      NULL);
		gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (chooser), FALSE);
		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (chooser),
						     g_get_home_dir ());
	
		gtk_widget_show (chooser);
		answer = gtk_dialog_run (GTK_DIALOG (chooser));
	
		if (answer == GTK_RESPONSE_OK)
			path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));
		else
			path = NULL;
	
		gtk_widget_destroy (chooser);
	}
	else
		path = g_strdup (name);

	if (!path || *path =='\0')
		return BONFIRE_DISC_CANCELLED;

	if (!bonfire_project_open_project_xml (project, path, &track)) {
		g_free (path);
		return BONFIRE_DISC_CANCELLED;
	}

	bonfire_project_size_set_size (BONFIRE_PROJECT_SIZE (project->priv->size_display),
				       0.0,
				       NULL);

	if (track->type == BONFIRE_DISC_TRACK_AUDIO) {
		bonfire_project_switch (project, BONFIRE_PROJECT_TYPE_AUDIO);
		disc = BONFIRE_DISC (project->priv->audio);

		gtk_window_set_title (GTK_WINDOW (toplevel), _("Bonfire - Audio disc"));
	}
	else if (track->type == BONFIRE_DISC_TRACK_DATA) {
		bonfire_project_switch (project, BONFIRE_PROJECT_TYPE_DATA);
		disc = BONFIRE_DISC (project->priv->data);

		gtk_window_set_title (GTK_WINDOW (toplevel), _("Bonfire - Data disc"));
	}
	else 
		disc = NULL;

	project->priv->project = path;

	result = bonfire_disc_load_track (disc, track);
	bonfire_track_free (track);

	return result;
}

/******************************** save project *********************************/
static void
bonfire_project_not_saved_dialog (BonfireProject *project)
{
	GtkWidget *dialog;
	xmlErrorPtr error;
	GtkWidget *toplevel;

	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	dialog = gtk_message_dialog_new (GTK_WINDOW (toplevel),
					 GTK_DIALOG_DESTROY_WITH_PARENT|
					 GTK_DIALOG_MODAL,
					 GTK_MESSAGE_WARNING,
					 GTK_BUTTONS_CLOSE,
					 _("Your project has not been saved:"));

	gtk_window_set_title (GTK_WINDOW (dialog), _("Unsaved project"));

	error = xmlGetLastError ();
	if (error)
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
							  error->message);
	else
		gtk_message_dialog_format_secondary_text (GTK_MESSAGE_DIALOG (dialog),
							  _("Unknown error."));	
	xmlResetLastError ();

	gtk_dialog_run (GTK_DIALOG (dialog));
	gtk_widget_destroy (dialog);
}

static gboolean
_save_audio_track_xml (xmlTextWriter *project,
		       BonfireDiscTrack *track)
{
	GSList *iter;
	int success;
	char *uri;

	for (iter = track->contents.uris; iter; iter = iter->next) {
		uri = iter->data;
		success = xmlTextWriterWriteElement (project, (xmlChar *) "uri", (xmlChar *) uri);
		if (success == -1)
			return FALSE;
	}

	return TRUE;
}

static gboolean
_save_data_track_xml (xmlTextWriter *project,
		      BonfireDiscTrack *track)
{
	char *uri;
	int success;
	GSList *iter;
	GSList *grafts;
	BonfireGraftPt *graft;

	for (grafts = track->contents.data.grafts; grafts; grafts = grafts->next) {
		graft = grafts->data;

		success = xmlTextWriterStartElement (project, (xmlChar *) "graft");
		if (success < 0)
			return FALSE;

		success = xmlTextWriterWriteElement (project, (xmlChar *) "path", (xmlChar *) graft->path);
		if (success < 0)
			return FALSE;

		if (graft->uri) {
			success = xmlTextWriterWriteElement (project, (xmlChar *) "uri", (xmlChar *) graft->uri);
			if (success < 0)
				return FALSE;
		}

		for (iter = graft->excluded; iter; iter = iter->next) {
			uri = iter->data;
			success = xmlTextWriterWriteElement (project, (xmlChar *) "excluded", (xmlChar *) uri);
			if (success < 0)
				return FALSE;
		}

		success = xmlTextWriterEndElement (project); /* graft */
		if (success < 0)
			return FALSE;
	}

	/* save restored uris */
	for (iter = track->contents.data.restored; iter; iter = iter->next) {
		uri = iter->data;
		success = xmlTextWriterWriteElement (project, (xmlChar *) "restored", (xmlChar *) uri);
		if (success < 0)
			return FALSE;
	}

	/* NOTE: we don't write symlinks and unreadable they are useless */
	return TRUE;
}

static gboolean 
bonfire_project_save_project_xml (BonfireProject *proj,
				  const char *path,
				  BonfireDiscTrack *track)
{
	xmlTextWriter *project;
	gboolean retval;
	int success;

	project = xmlNewTextWriterFilename (path, 0);
	if (!project) {
		bonfire_project_not_saved_dialog (proj);
		return FALSE;
	}

	xmlTextWriterSetIndent (project, 1);
	xmlTextWriterSetIndentString (project, (xmlChar *) "\t");

	success = xmlTextWriterStartDocument (project,
					      NULL,
					      NULL,
					      NULL);
	if (success < 0)
		goto error;

	success = xmlTextWriterStartElement (project, (xmlChar *) "bonfireproject");
	if (success < 0)
		goto error;

	/* write the name of the version */
	success = xmlTextWriterWriteElement (project,
					     (xmlChar *) "version",
					     (xmlChar *) BONFIRE_PROJECT_VERSION);
	if (success < 0)
		goto error;

	success = xmlTextWriterStartElement (project, (xmlChar *) "track");
	if (success < 0)
		goto error;

	if (track->type == BONFIRE_DISC_TRACK_AUDIO) {
		success = xmlTextWriterStartElement (project, (xmlChar *) "audio");
		if (success < 0)
			goto error;

		retval = _save_audio_track_xml (project, track);
		if (!retval)
			goto error;

		success = xmlTextWriterEndElement (project); /* audio */
		if (success < 0)
			goto error;
	}
	else if (track->type == BONFIRE_DISC_TRACK_DATA) {
		success = xmlTextWriterStartElement (project, (xmlChar *) "data");
		if (success < 0)
			goto error;

		retval = _save_data_track_xml (project, track);
		if (!retval)
			goto error;

		success = xmlTextWriterEndElement (project); /* data */
		if (success < 0)
			goto error;
	}
	else 
		retval = FALSE;

	success = xmlTextWriterEndElement (project); /* track */
	if (success < 0)
		goto error;

	success = xmlTextWriterEndElement (project); /* bonfireproject */
	if (success < 0)
		goto error;

	xmlTextWriterEndDocument (project);
	xmlFreeTextWriter (project);
	return TRUE;

error:
	xmlTextWriterEndDocument (project);
	xmlFreeTextWriter (project);
	g_remove (path);
	
	bonfire_project_not_saved_dialog (proj);
	return FALSE;
}

static gboolean
bonfire_project_save_project_real (BonfireProject *project,
				   char *path)
{
	BonfireDiscResult result;
	BonfireDiscTrack *track;
	BonfireDisc *disc;
	gboolean retval;

	track = g_new0 (BonfireDiscTrack, 1);
	switch (project->priv->type) {
	case BONFIRE_PROJECT_TYPE_AUDIO:
		disc = BONFIRE_DISC (project->priv->audio);
		track->type = BONFIRE_DISC_TRACK_AUDIO;
		break;
	case BONFIRE_PROJECT_TYPE_DATA:
		disc = BONFIRE_DISC (project->priv->data);
		track->type = BONFIRE_DISC_TRACK_DATA;
		break;
	default:
		return FALSE;
	}

	result = bonfire_project_check_status (project, disc);
	if (result != BONFIRE_DISC_OK)
		return FALSE;

	result = bonfire_disc_get_track (disc, track);
	if (result == BONFIRE_DISC_ERROR_EMPTY_SELECTION) {
		if (BONFIRE_IS_AUDIO_DISC (disc))
			bonfire_project_no_song_dialog (project);
		else if (BONFIRE_IS_DATA_DISC (disc))
			bonfire_project_no_file_dialog (project);

		if (path)
			g_free (path);
		return FALSE;
	}
	else if (result != BONFIRE_DISC_OK) {
		bonfire_project_not_saved_dialog (project);
		return FALSE;
	}

	retval = bonfire_project_save_project_xml (project,
						   path ? path : project->priv->project,
						   track);

	if (retval) {
		char *title;
		GtkWidget *toplevel;

		if (path) {
			if (project->priv->project)
				g_free (project->priv->project);
			project->priv->project = path;
		}

		if (track->type == BONFIRE_DISC_TRACK_DATA)
			title = g_strdup_printf (_("Bonfire - data disc"));
		else
			title = g_strdup (_("Bonfire - audio disc"));

#ifdef BUILD_GDL
		GdlDock *dock;
		dock = gdl_dock_object_get_toplevel (GDL_DOCK_OBJECT (GTK_WIDGET (project)->parent));
		if (dock)
			toplevel = gtk_widget_get_toplevel (GTK_WIDGET (dock));
		else
			toplevel = NULL;
#else
		toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
#endif	
		gtk_window_set_title (GTK_WINDOW (toplevel), title);
		g_free (title);
	}
	else if (path)
		g_free (path);

	bonfire_track_free (track);
	return retval;
}

static char *
bonfire_project_save_project_ask_for_path (BonfireProject *project)
{
	GtkWidget *toplevel;
	GtkWidget *chooser;
	int answer;
	char *path;

	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (project));
	chooser = gtk_file_chooser_dialog_new (_("Save current project"),
					       GTK_WINDOW (toplevel),
					       GTK_FILE_CHOOSER_ACTION_SAVE,
					       GTK_STOCK_OK, GTK_RESPONSE_OK,
					       GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					       NULL);
	gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (chooser), FALSE);
	gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (chooser),
					     g_get_home_dir ());

	gtk_widget_show (chooser);
	answer = gtk_dialog_run (GTK_DIALOG (chooser));
	if (answer == GTK_RESPONSE_OK) {
		path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (chooser));

		if (*path == '\0') {
			g_free (path);
			path = NULL;
		}
	}
	else
		path = NULL;
	gtk_widget_destroy (chooser);

	return path;
}

gboolean
bonfire_project_save_project (BonfireProject *project)
{
	char *path = NULL;

	if (!project->priv->project
	&&  !(path = bonfire_project_save_project_ask_for_path (project)))
		return FALSE;

	return bonfire_project_save_project_real (project, path);
}

gboolean
bonfire_project_save_project_as (BonfireProject *project)
{
	char *path;

	if (!(path = bonfire_project_save_project_ask_for_path (project)))
		return FALSE;

	return bonfire_project_save_project_real (project, path);
}
