/***************************************************************************
*            recorder-selection.c
*
*  mer jun 15 12:40:07 2005
*  Copyright  2005  Philippe Rouquier
*  bonfire-app@wanadoo.fr
****************************************************************************/

/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif


#include <glib.h>
#include <glib/gi18n-lib.h>

#include <gtk/gtkbutton.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkmessagedialog.h>
#include <gtk/gtkdialog.h>

#include <nautilus-burn-drive.h>
#include <nautilus-burn-drive-selection.h>
#include <nautilus-burn-recorder.h>

#include "burn-caps.h"
#include "recorder-selection.h"
#include "utils.h"
#include "burn.h"

#define ICON_SIZE 48

static void bonfire_recorder_selection_class_init (BonfireRecorderSelectionClass *klass);
static void bonfire_recorder_selection_init (BonfireRecorderSelection *sp);
static void bonfire_recorder_selection_finalize (GObject * object);

static void bonfire_recorder_selection_set_property (GObject *object,
						     guint property_id,
						     const GValue *value,
						     GParamSpec *pspec);
static void bonfire_recorder_selection_get_property (GObject *object,
						     guint property_id,
						     GValue *value,
						     GParamSpec *pspec);

static void bonfire_recorder_selection_button_cb (GtkWidget *button,
						   BonfireRecorderSelection *selection);

static void bonfire_recorder_selection_drive_changed_cb (NautilusBurnDriveSelection *selector, 
							  const char *device,
							  BonfireRecorderSelection *selection);

enum {
	PROP_NONE,
	PROP_IMAGE,
	PROP_SHOW_PROPS,
	PROP_SHOW_RECORDERS_ONLY
};

enum {
	DEVICE_CHANGED_SIGNAL,
	LAST_SIGNAL
};
static guint bonfire_recorder_selection_signals [LAST_SIGNAL] = { 0 };


struct BonfireRecorderSelectionPrivate {
	char *image_path;
	GtkWidget *selection;
	GHashTable *settings;
	GtkWidget *dialog;
	GtkWidget *infos;
	GtkWidget *image;
	GtkWidget *props;
	NautilusBurnDrive *drive;

	gint added_signal;
	gint removed_signal;

	BonfireBurnCaps *caps;
	NautilusBurnMediaType media_type;
	BonfireTrackSourceType track_type;
};

static GObjectClass *parent_class = NULL;

GType
bonfire_recorder_selection_get_type ()
{
	static GType type = 0;

	if (type == 0) {
		static const GTypeInfo our_info = {
			sizeof (BonfireRecorderSelectionClass),
			NULL,
			NULL,
			(GClassInitFunc)
			    bonfire_recorder_selection_class_init,
			NULL,
			NULL,
			sizeof (BonfireRecorderSelection),
			0,
			(GInstanceInitFunc)
			    bonfire_recorder_selection_init,
		};

		type = g_type_register_static (GTK_TYPE_VBOX,
					       "BonfireRecorderSelection",
					       &our_info, 0);
	}

	return type;
}

static void
bonfire_recorder_selection_class_init (BonfireRecorderSelectionClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	parent_class = g_type_class_peek_parent (klass);
	object_class->finalize = bonfire_recorder_selection_finalize;
	object_class->set_property = bonfire_recorder_selection_set_property;
	object_class->get_property = bonfire_recorder_selection_get_property;

	bonfire_recorder_selection_signals [DEVICE_CHANGED_SIGNAL] =
	    g_signal_new ("device-changed",
			  G_TYPE_FROM_CLASS (klass),
			  G_SIGNAL_RUN_LAST,
			  G_STRUCT_OFFSET (BonfireRecorderSelectionClass,
					   device_changed),
			  NULL, NULL,
			  g_cclosure_marshal_VOID__BOOLEAN,
			  G_TYPE_NONE,
			  1,
			  G_TYPE_BOOLEAN);

	g_object_class_install_property (object_class,
					 PROP_IMAGE,
					 g_param_spec_boolean ("file-image", NULL, NULL,
							       FALSE, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_SHOW_PROPS,
					 g_param_spec_boolean ("show-properties", NULL, NULL,
							       TRUE, G_PARAM_READWRITE));
	g_object_class_install_property (object_class,
					 PROP_SHOW_RECORDERS_ONLY,
					 g_param_spec_boolean ("show-recorders-only", NULL, NULL,
							       TRUE, G_PARAM_READWRITE));
}

static void
bonfire_recorder_selection_create_prop_button (BonfireRecorderSelection *selection)
{
	GtkWidget *parent;

	selection->priv->props = gtk_button_new_from_stock (GTK_STOCK_PROPERTIES);
	g_signal_connect (G_OBJECT (selection->priv->props),
			  "clicked",
			  G_CALLBACK (bonfire_recorder_selection_button_cb),
			  selection);

	parent = gtk_widget_get_parent (selection->priv->selection);
	gtk_box_pack_start (GTK_BOX (parent), selection->priv->props, FALSE, FALSE, 0);
}

static void
bonfire_recorder_selection_set_property (GObject *object,
					 guint property_id,
					 const GValue *value,
					 GParamSpec *pspec)
{
	BonfireRecorderSelection *selection;

	selection = BONFIRE_RECORDER_SELECTION (object);
	switch (property_id) {
	case PROP_IMAGE:
		g_object_set_property (G_OBJECT (selection->priv->selection),
				       "file-image",
				       value);
		break;
	case PROP_SHOW_PROPS:
		if (g_value_get_boolean (value))  {
			if (!selection->priv->props) {
				bonfire_recorder_selection_create_prop_button (selection);
				gtk_widget_show (selection->priv->props);
			}
		}
		else {
			gtk_widget_destroy (selection->priv->props);
			selection->priv->props = NULL;
		}
		break;
	case PROP_SHOW_RECORDERS_ONLY:
		g_object_set_property (G_OBJECT (selection->priv->selection),
				       "show-recorders-only",
				       value);
		break;

	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void
bonfire_recorder_selection_get_property (GObject *object,
					 guint property_id,
					 GValue *value,
					 GParamSpec *pspec)
{
	BonfireRecorderSelection *selection;

	selection = BONFIRE_RECORDER_SELECTION (object);
	switch (property_id) {
	case PROP_IMAGE:
		g_object_get_property (G_OBJECT (selection->priv->selection),
				       "file-image",
				       value);
		break;
	case PROP_SHOW_PROPS:
		g_value_set_boolean (value, GTK_WIDGET_VISIBLE (selection->priv->props));
		break;
	case PROP_SHOW_RECORDERS_ONLY:
		g_object_get_property (G_OBJECT (selection->priv->selection),
				       "show-recorders-only",
				       value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
	}
}

static void
bonfire_recorder_selection_init (BonfireRecorderSelection *obj)
{
	GtkWidget *box;

	obj->priv = g_new0 (BonfireRecorderSelectionPrivate, 1);
	gtk_box_set_spacing (GTK_BOX (obj), 12);

	obj->priv->caps = bonfire_burn_caps_get_default ();

	obj->priv->settings = g_hash_table_new_full (g_str_hash,
						     g_str_equal,
						     g_free,
						     g_free);

	box = gtk_hbox_new (FALSE, 12);
	obj->priv->selection = nautilus_burn_drive_selection_new ();

	gtk_box_pack_start (GTK_BOX (box),
			    obj->priv->selection,
			    FALSE,
			    FALSE,
			    0);

	bonfire_recorder_selection_create_prop_button (obj);
	gtk_box_pack_start (GTK_BOX (obj), box, FALSE, FALSE, 0);

	box = gtk_hbox_new (FALSE, 12);
	obj->priv->image = gtk_image_new ();
	obj->priv->infos = gtk_label_new ("");
	gtk_misc_set_alignment (GTK_MISC (obj->priv->infos), 0.0, 0.0);

	gtk_box_pack_start (GTK_BOX (box), obj->priv->image, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (box), obj->priv->infos, FALSE, FALSE, 0);
	gtk_box_pack_start (GTK_BOX (obj), box, FALSE, FALSE, 0);

	g_signal_connect (obj->priv->selection,
			  "device-changed",
			  G_CALLBACK (bonfire_recorder_selection_drive_changed_cb),
			  obj);
}

static void
bonfire_recorder_selection_finalize (GObject *object)
{
	BonfireRecorderSelection *cobj;

	cobj = BONFIRE_RECORDER_SELECTION (object);

	if (cobj->priv->image_path) {
		g_free (cobj->priv->image_path);
		cobj->priv->image_path = NULL;
	}

	if (cobj->priv->added_signal) {
		g_signal_handler_disconnect (cobj->priv->drive,
					     cobj->priv->added_signal);
		cobj->priv->added_signal = 0;
	}

	if (cobj->priv->removed_signal) {
		g_signal_handler_disconnect (cobj->priv->drive,
					     cobj->priv->removed_signal);
		cobj->priv->removed_signal = 0;
	}

	if (cobj->priv->caps) {
		g_object_unref (cobj->priv->caps);
		cobj->priv->caps = NULL;
	}

	g_hash_table_destroy (cobj->priv->settings);

	g_free (cobj->priv);
	G_OBJECT_CLASS (parent_class)->finalize (object);
}

static char *
_get_new_image_path (void)
{
	char *path;
	int i = 0;

	path = g_strdup_printf ("%s/bonfire.iso",
				g_get_home_dir ());

	while (g_file_test (path, G_FILE_TEST_EXISTS)) {
		g_free (path);

		path = g_strdup_printf ("%s/bonfire-%i.iso",
					g_get_home_dir (),
					i);
		i ++;
	};
	return path;
}

static gboolean
bonfire_recorder_selection_update_info (BonfireRecorderSelection *selection,
					NautilusBurnMediaType type,
					gboolean is_rewritable,
					gboolean has_audio,
					gboolean has_data,
					gboolean is_blank)
{
	char *info;
	GdkPixbuf *pixbuf = NULL;
	gboolean can_record = FALSE;
	char *types [] = { 	NULL,
				NULL,
				NULL,
				"gnome-dev-cdrom",
				"gnome-dev-disc-cdr",
				"gnome-dev-disc-cdrw",
				"gnome-dev-disc-dvdrom",
				"gnome-dev-disc-dvdr",
				"gnome-dev-disc-dvdrw",
				"gnome-dev-disc-dvdram",
				"gnome-dev-disc-dvdr-plus",
				"gnome-dev-disc-dvdrw", /* FIXME */
				"gnome-dev-disc-dvdr-plus" /* FIXME */,
				NULL };

	selection->priv->media_type = type;

	if (type == NAUTILUS_BURN_MEDIA_TYPE_BUSY) {
		info = g_strdup (_("<i>The disc is mounted.</i>"));
		can_record = TRUE;
	}
	else if (type == NAUTILUS_BURN_MEDIA_TYPE_ERROR) {
		info = g_strdup (_("<i>There is no disc in the drive.</i>"));
	}
	else if (type == NAUTILUS_BURN_MEDIA_TYPE_UNKNOWN) {
		info = g_strdup (_("<i>Unknown type of disc.</i>"));
	}
	else if (!is_blank && !is_rewritable) {
		info = g_strdup_printf (_("The <b>%s</b> is not writable."),
					nautilus_burn_drive_media_type_get_string (type));
		pixbuf = bonfire_utils_get_icon (types [type], ICON_SIZE);
	}
	else if (has_data) {
		info = g_strdup_printf (_("The <b>%s</b> is ready.\nIt contains data."),
					nautilus_burn_drive_media_type_get_string (type));
		pixbuf = bonfire_utils_get_icon (types [type], ICON_SIZE);
		can_record = TRUE;
	}
	else if (has_audio) {
		info = g_strdup_printf (_("The <b>%s</b> is ready.\nIt contains audio tracks."),
					nautilus_burn_drive_media_type_get_string (type));
		pixbuf = bonfire_utils_get_icon (types [type], ICON_SIZE);
		can_record = TRUE;
	}
	else {
		info = g_strdup_printf (_("The <b>%s</b> is ready.\nIt is empty."),
					nautilus_burn_drive_media_type_get_string (type));
		pixbuf = bonfire_utils_get_icon (types [type], ICON_SIZE);
		can_record = TRUE;
	}

	if (selection->priv->props) {
		if (can_record)
			gtk_widget_set_sensitive (selection->priv->props, TRUE);
		else
			gtk_widget_set_sensitive (selection->priv->props, FALSE);
	}

	gtk_label_set_markup (GTK_LABEL (selection->priv->infos), info);
	if (!pixbuf)
		pixbuf = bonfire_utils_get_icon ("gnome-dev-removable", ICON_SIZE);

	gtk_image_set_from_pixbuf (GTK_IMAGE (selection->priv->image), pixbuf);
	g_object_unref (pixbuf);
	g_free (info);

	return can_record;
}

static void
bonfire_recorder_selection_drive_media_added_cb (NautilusBurnDrive *drive,
						 BonfireRecorderSelection *selection)
{
	gboolean can_record;
	NautilusBurnMediaType type;
	gboolean is_rewritable, has_audio, has_data, is_blank;

	type = nautilus_burn_drive_get_media_type_full (drive,
							&is_blank,
							&is_rewritable,
							&has_data,
							&has_audio);

	can_record = bonfire_recorder_selection_update_info (selection,
							     type,
							     is_rewritable,
							     has_audio,
							     has_data,
							     is_blank);
	g_signal_emit (selection,
		       bonfire_recorder_selection_signals [DEVICE_CHANGED_SIGNAL],
		       0,
		       can_record);
}

static void
bonfire_recorder_selection_drive_media_removed_cb (NautilusBurnDrive *drive,
						   BonfireRecorderSelection *selection)
{
	gboolean can_record;

	if (selection->priv->dialog)
		gtk_dialog_response (GTK_DIALOG (selection->priv->dialog),
				     GTK_RESPONSE_CANCEL);

	/* we don't look at the drive contents since we already
	 * know that there is nothing inside. If we did it could
	 * force the drive to reload the disc */
	can_record = bonfire_recorder_selection_update_info (selection,
							     NAUTILUS_BURN_MEDIA_TYPE_ERROR,
							     FALSE,
							     FALSE,
							     FALSE,
							     FALSE);
	g_signal_emit (selection,
		       bonfire_recorder_selection_signals [DEVICE_CHANGED_SIGNAL],
		       0,
		       can_record);
}

static void
bonfire_recorder_selection_update_drive_info (BonfireRecorderSelection *selection)
{
	guint added_signal = 0;
	guint removed_signal = 0;
	NautilusBurnDrive *drive;
	gboolean can_record = FALSE;
	NautilusBurnMediaType type;
	gboolean is_rewritable, has_audio, has_data, is_blank;

	drive = nautilus_burn_drive_selection_get_active (NAUTILUS_BURN_DRIVE_SELECTION (selection->priv->selection));
	if (drive == NULL) {
		GdkPixbuf *pixbuf;

		gtk_widget_set_sensitive (selection->priv->selection, FALSE);
		gtk_label_set_markup (GTK_LABEL (selection->priv->infos),
				      _("<b>There is no available drive.</b>"));

		pixbuf = bonfire_utils_get_icon ("gnome-dev-removable", ICON_SIZE);
		gtk_image_set_from_pixbuf (GTK_IMAGE (selection->priv->image), pixbuf);
		g_object_unref (pixbuf);
		goto end;
	}

	if (selection->priv->drive
	&&  nautilus_burn_drive_equal (selection->priv->drive, drive)) {
		nautilus_burn_drive_unref (drive);
		return;
	}

	if (drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE) {
		GdkPixbuf *pixbuf;
		char *info;

		if (!selection->priv->image_path) {
			char *path;

			path = _get_new_image_path ();
			info = g_strdup_printf (_("The <b>image</b> will be saved to\n%s"),
						path);
			g_free (path);
		}
		else
			info = g_strdup_printf (_("The <b>image</b> will be saved to\n%s"),
						selection->priv->image_path);

		gtk_label_set_markup (GTK_LABEL (selection->priv->infos), info);
		g_free (info);

		pixbuf = bonfire_utils_get_icon_for_mime ("application/x-cd-image", ICON_SIZE);
		gtk_image_set_from_pixbuf (GTK_IMAGE (selection->priv->image), pixbuf);
		g_object_unref (pixbuf);

		can_record = TRUE;
		goto end;
	}

	type = nautilus_burn_drive_get_media_type_full (drive,
							&is_blank,
							&is_rewritable,
							&has_data,
							&has_audio);

	can_record = bonfire_recorder_selection_update_info (selection,
							     type,
							     is_rewritable,
							     has_audio,
							     has_data,
							     is_blank);

	g_object_set (G_OBJECT (drive), "enable-monitor", TRUE, NULL);
	added_signal = g_signal_connect (G_OBJECT (drive),
					 "media-added",
					 G_CALLBACK (bonfire_recorder_selection_drive_media_added_cb),
					 selection);
	removed_signal = g_signal_connect (G_OBJECT (drive),
					   "media-removed",
					   G_CALLBACK (bonfire_recorder_selection_drive_media_removed_cb),
					   selection);

end:

	if (selection->priv->added_signal) {
		g_signal_handler_disconnect (selection->priv->drive,
					     selection->priv->added_signal);
		selection->priv->added_signal = 0;
	}

	if (selection->priv->removed_signal) {
		g_signal_handler_disconnect (selection->priv->drive,
					     selection->priv->removed_signal);
		selection->priv->removed_signal = 0;
	}

	if (selection->priv->drive) {
		g_object_set (G_OBJECT (selection->priv->drive),
			      "enable-monitor", FALSE,
			      NULL);
		nautilus_burn_drive_unref (selection->priv->drive);
	}

	selection->priv->drive = drive;
	selection->priv->added_signal = added_signal;
	selection->priv->removed_signal = removed_signal;
	g_signal_emit (selection,
		       bonfire_recorder_selection_signals [DEVICE_CHANGED_SIGNAL],
		       0,
		       can_record);
}

static void
bonfire_recorder_selection_drive_changed_cb (NautilusBurnDriveSelection *selector,
					     const char *device,
					     BonfireRecorderSelection *selection)
{
	bonfire_recorder_selection_update_drive_info (selection);
}

GtkWidget *
bonfire_recorder_selection_new (BonfireTrackSourceType type)
{
	BonfireRecorderSelection *obj;

	obj = BONFIRE_RECORDER_SELECTION (g_object_new (BONFIRE_TYPE_RECORDER_SELECTION,
					                NULL));
	bonfire_recorder_selection_update_drive_info (obj);
	return GTK_WIDGET (obj);
}

void
bonfire_recorder_selection_set_track_type (BonfireRecorderSelection *selection,
					   BonfireTrackSourceType type)
{
	selection->priv->track_type = type;
}

static gboolean
bonfire_recorder_selection_set_default_setting_for_drive (BonfireRecorderSelection *selection,
							  BonfireDriveProp *settings)
{
	BonfireBurnFlag default_flags;

	if (bonfire_burn_caps_get_default_flags (selection->priv->caps,
						 selection->priv->track_type,
						 selection->priv->drive ? selection->priv->drive->type:0,
						 selection->priv->media_type,
						 &default_flags) != BONFIRE_BURN_OK)
		return FALSE;

	settings->drive.speed = selection->priv->drive->max_speed_write;
	settings->drive.flags = default_flags;

	return TRUE;
}

static void
bonfire_recorder_selection_drive_properties (BonfireRecorderSelection *selection)
{
	NautilusBurnDrive *drive;
	NautilusBurnMediaType media;
	BonfireDriveProp *prop;
	BonfireBurnFlag flags = 0;
	GtkWidget *combo;
	GtkWidget *toplevel;
	GtkWidget *toggle_otf = NULL;
	GtkWidget *toggle_simulation = NULL;
	GtkWidget *toggle_eject = NULL;
	GtkWidget *toggle_burnproof = NULL;
	GtkWidget *dialog;
	char *header, *text;
	gint result, i;
	gint speed;
	GSList *list = NULL;

	/* */
	nautilus_burn_drive_ref (selection->priv->drive);
	drive = selection->priv->drive;

	/* dialog */
	header = g_strdup_printf (_("Properties of %s"),
				  drive->display_name);
	toplevel = gtk_widget_get_toplevel (GTK_WIDGET (selection));
	dialog = gtk_dialog_new_with_buttons (header,
					      GTK_WINDOW (toplevel),
					      GTK_DIALOG_DESTROY_WITH_PARENT |
					      GTK_DIALOG_DESTROY_WITH_PARENT,
					      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
					      GTK_STOCK_OK, GTK_RESPONSE_ACCEPT,
					      NULL);
	gtk_window_set_default_size (GTK_WINDOW (dialog), 340, 250);
	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
	g_free (header);

	/* Speed combo */
	media = nautilus_burn_drive_get_media_type (drive);
	speed = nautilus_burn_drive_get_max_speed_write (drive);

	combo = gtk_combo_box_new_text ();
	gtk_combo_box_append_text (GTK_COMBO_BOX (combo), _("Max speed"));

	/* FIXME : we should use nautilus_burn_drive_get_speeds in the future
	 * when it actually works */
	if (media > NAUTILUS_BURN_MEDIA_TYPE_CDRW)
		for (i = 2; i < speed; i *= 2) {
		/* FIXME : this has changed in 2.15/16 speeds are given in bytes */
		text = g_strdup_printf ("%i x (DVD)", i);
		gtk_combo_box_append_text (GTK_COMBO_BOX (combo), text);
		g_free (text);
	}
	else for (i = 2; i < speed;i += 2) {
		/* FIXME : this has changed in 2.15/16 speeds are given in bytes */
		text = g_strdup_printf ("%i x (CD)", i);
		gtk_combo_box_append_text (GTK_COMBO_BOX (combo), text);
		g_free (text);
	}

	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
			    bonfire_utils_pack_properties (_("<b>Burning speed</b>"),
							   combo, NULL),
			    FALSE, FALSE, 0);

	prop = g_hash_table_lookup (selection->priv->settings,
				    drive->display_name); /* FIXME what about drives with the same display names */

	if (!prop) {
		prop = g_new0 (BonfireDriveProp, 1);
		bonfire_recorder_selection_set_default_setting_for_drive (selection, prop);
		g_hash_table_insert (selection->priv->settings,
				     g_strdup (drive->display_name),
				     prop);
	}

	if (prop->drive.speed == 0
	||  prop->drive.speed >= drive->max_speed_write)
		gtk_combo_box_set_active (GTK_COMBO_BOX (combo), 0);
	else
		gtk_combo_box_set_active (GTK_COMBO_BOX (combo),
					  prop->drive.speed / 2);

	/* properties */
	bonfire_burn_caps_get_supported_flags (selection->priv->caps,
					       selection->priv->track_type,
					       selection->priv->drive->type,
					       selection->priv->media_type,
					       &flags);

	if (flags & BONFIRE_BURN_FLAG_DUMMY) {
		toggle_simulation = gtk_check_button_new_with_label (_("Simulate the burning"));
		if (prop->drive.flags & BONFIRE_BURN_FLAG_DUMMY)
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle_simulation), TRUE);

		list = g_slist_prepend (list, toggle_simulation);
	}

	if (flags & BONFIRE_BURN_FLAG_EJECT) {
		toggle_eject = gtk_check_button_new_with_label (_("Eject after burning"));
		if (prop->drive.flags & BONFIRE_BURN_FLAG_EJECT)
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle_eject), TRUE);

		list = g_slist_prepend (list, toggle_eject);
	}

	if (flags & BONFIRE_BURN_FLAG_BURNPROOF) {
		toggle_burnproof = gtk_check_button_new_with_label (_("Use burnproof (decrease the risk of failures)"));
		if (prop->drive.flags & BONFIRE_BURN_FLAG_BURNPROOF)
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle_burnproof), TRUE);

		list = g_slist_prepend (list, toggle_burnproof);
	}

	if (flags & BONFIRE_BURN_FLAG_ON_THE_FLY) {
		toggle_otf = gtk_check_button_new_with_label (_("Burn the image directly without saving it to disc"));
		if (prop->drive.flags & BONFIRE_BURN_FLAG_ON_THE_FLY)
			gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (toggle_otf), TRUE);

		if (media > NAUTILUS_BURN_MEDIA_TYPE_CDRW
		&&  selection->priv->track_type != BONFIRE_TRACK_SOURCE_DISC)
			gtk_widget_set_sensitive (toggle_otf, FALSE);

		list = g_slist_prepend (list, toggle_otf);
	}

	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
			    bonfire_utils_pack_properties_list (_("<b>Options</b>"), list),
			    FALSE,
			    FALSE, 0);

	g_slist_free (list);

	gtk_widget_show_all (dialog);
	selection->priv->dialog = dialog;
	result = gtk_dialog_run (GTK_DIALOG (dialog));
	selection->priv->dialog = NULL;

	if (result != GTK_RESPONSE_ACCEPT) {
		nautilus_burn_drive_unref (drive);
		gtk_widget_destroy (dialog);
		return;
	}

	prop->drive.speed = gtk_combo_box_get_active (GTK_COMBO_BOX (combo));
	if (prop->drive.speed == 0)
		prop->drive.speed = drive->max_speed_write;
	else
		prop->drive.speed = prop->drive.speed * 2;

	flags = prop->drive.flags;

	if (toggle_otf
	&&  gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (toggle_otf)) == TRUE)
		flags |= BONFIRE_BURN_FLAG_ON_THE_FLY;
	else
		flags &= ~BONFIRE_BURN_FLAG_ON_THE_FLY;

	if (toggle_eject
	&&  gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (toggle_eject)) == TRUE)
		flags |= BONFIRE_BURN_FLAG_EJECT;
	else
		flags &= ~BONFIRE_BURN_FLAG_EJECT;

	if (toggle_simulation
	&&  gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (toggle_simulation)) == TRUE)
		flags |= BONFIRE_BURN_FLAG_DUMMY;
	else
		flags &= ~BONFIRE_BURN_FLAG_DUMMY;

	if (toggle_burnproof
	&&  gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (toggle_burnproof)) == TRUE)
		flags |= BONFIRE_BURN_FLAG_BURNPROOF;
	else
		flags &= ~BONFIRE_BURN_FLAG_BURNPROOF;

	prop->drive.flags = flags;

	gtk_widget_destroy (dialog);
	nautilus_burn_drive_unref (drive);
}

static void
bonfire_recorder_selection_button_cb (GtkWidget * button,
				      BonfireRecorderSelection * selection)
{

	if (selection->priv->drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE) {
		GtkWidget *toplevel;
		GtkWidget *dialog;
		int answer;

		toplevel = gtk_widget_get_toplevel (GTK_WIDGET (selection));
		dialog = gtk_file_chooser_dialog_new (_("Choose a location for the disc image"),
						      GTK_WINDOW (toplevel),
						      GTK_FILE_CHOOSER_ACTION_SAVE,
						      GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
						      GTK_STOCK_SAVE, GTK_RESPONSE_OK,
						      NULL);

		gtk_file_chooser_set_do_overwrite_confirmation (GTK_FILE_CHOOSER (dialog), TRUE);
		gtk_file_chooser_set_local_only (GTK_FILE_CHOOSER (dialog), TRUE);
		gtk_file_chooser_set_current_folder (GTK_FILE_CHOOSER (dialog),
						     g_get_home_dir ());
	
		gtk_widget_show (dialog);
		answer = gtk_dialog_run (GTK_DIALOG (dialog));

		if (answer != GTK_RESPONSE_OK) {
			gtk_widget_destroy (dialog);
			return;
		}

		if (selection->priv->image_path)
			g_free (selection->priv->image_path);

		selection->priv->image_path = gtk_file_chooser_get_filename (GTK_FILE_CHOOSER (dialog));
		gtk_widget_destroy (dialog);
	}
	else
		bonfire_recorder_selection_drive_properties (selection);
}

void
bonfire_recorder_selection_get_drive (BonfireRecorderSelection *selection,
				      NautilusBurnDrive **drive,
				      BonfireDriveProp *props)
{
	BonfireDriveProp *setting;

	g_return_if_fail (drive != NULL);

	if (!selection->priv->drive) {
		*drive = NULL;
		if (props) {
			props->drive.flags = BONFIRE_BURN_FLAG_NONE;
			props->drive.speed = 0;
		}
		return;
	}

	nautilus_burn_drive_ref (selection->priv->drive);
	*drive = selection->priv->drive;

	if (!props)
		return;

	if (selection->priv->drive->type == NAUTILUS_BURN_DRIVE_TYPE_FILE) {
		if (!selection->priv->image_path)
			props->output_path = _get_new_image_path ();
		else
			props->output_path = g_strdup (selection->priv->image_path);

		return;
	}

	setting = g_hash_table_lookup (selection->priv->settings,
				       selection->priv->drive->display_name);
	if (!setting) {
		bonfire_recorder_selection_set_default_setting_for_drive (selection, props);
		return;
	}

	props->drive.speed = setting->drive.speed;
	props->drive.flags = setting->drive.flags;
}
